import importlib.metadata
from contextlib import ExitStack
from typing import Optional
from typing import Union

import numpy
from ewoksdata.data.hdf5.dataset_writer import DatasetWriter
from ewoksdata.data.hdf5.dataset_writer import StackDatasetWriter
from ewoksdata.data.nexus import create_nexus_group
from ewoksdata.data.nexus import select_default_plot
from silx.io.dictdump import dicttonx

from .abstract import AbstractDataHandler
from .abstract import AbstractOutputHandler

version = importlib.metadata.version("ewoksfluo")


class _NexusDataHandler(AbstractDataHandler):
    def __init__(self, group: str, name: str, writer: DatasetWriter) -> None:
        self._writer = writer
        self._group = group
        self._name = name

    def add_points(
        self, value: numpy.ndarray, scan_index: Optional[int] = None
    ) -> None:
        self._writer.add_points(value)


class _NexusStackDataHandler(AbstractDataHandler):
    def __init__(self, group: str, name: str, writer: StackDatasetWriter) -> None:
        self._writer = writer
        self._group = group
        self._name = name

    def add_points(
        self, value: numpy.ndarray, stack_index: Optional[int] = None
    ) -> None:
        if stack_index is not None:
            self._writer.add_points(value, stack_index)


NexusDataHandlerType = Union[_NexusDataHandler, _NexusStackDataHandler]


class NexusOutputHandler(AbstractOutputHandler):
    def __init__(self, url: str, default_group: Optional[str] = None, **open_options):
        self._url = url
        self._open_options = open_options
        self._default_group = default_group

        # Output
        self._already_existed = False
        self._xrf_results_uri = None
        self._output_root_uri = None

        # Context parameters
        self._reset_context_parameters()

    def _reset_context_parameters(self):
        self._ctx_stack = None
        self._nxprocess = None
        self._results = None
        self._nxdata_groups = dict()

    @property
    def xrf_results_uri(self) -> Optional[str]:
        return self._xrf_results_uri

    @property
    def output_root_uri(self) -> Optional[str]:
        return self._output_root_uri

    @property
    def already_existed(self) -> bool:
        return self._already_existed

    def __enter__(self) -> "NexusOutputHandler":
        self._ctx_stack = ExitStack().__enter__()
        ctx = create_nexus_group(self._url, default_levels=("results", "fit"))
        self._nxprocess, self._already_existed = self._ctx_stack.enter_context(ctx)
        self._nxprocess.attrs["NX_class"] = "NXprocess"
        if not self._already_existed:
            self._nxprocess["program"] = "ewoksfluo"
            self._nxprocess["version"] = version
        self._xrf_results_uri = (
            f"{self._nxprocess.file.filename}::{self._nxprocess.name}/results"
        )
        entry_name = [s for s in self._nxprocess.parent.name.split("/") if s][0]
        self._output_root_uri = f"{self._nxprocess.file.filename}::/{entry_name}"
        return self

    def __exit__(self, *args) -> None:
        try:
            self._finalize()
        finally:
            try:
                result = self._ctx_stack.__exit__(*args)
            finally:
                self._reset_context_parameters()
        return result

    def create_group(self, name: str, data: dict) -> None:
        is_nxdata = data["@NX_class"] == "NXdata"
        if is_nxdata:
            if self._results is None:
                self._results = self._nxprocess.create_group("results")
                self._results.attrs["NX_class"] = "NXcollection"
            parent = self._results
        else:
            parent = self._nxprocess
        if name in parent:
            return
        dicttonx(data, parent.file, h5path=f"{parent.name}/{name}")
        if is_nxdata:
            group = parent[name]
            self._nxdata_groups[name] = {"group": group, "signals": list()}
            if name == self._default_group:
                select_default_plot(group)

    def create_nxdata_handler(
        self,
        group: str,
        name: str,
        npoints: int,
        attrs: Optional[dict] = None,
        stack_depth: Optional[int] = None,
    ) -> NexusDataHandlerType:
        name = name.replace(" ", "_")
        nxdata = self._nxdata_groups[group]
        if stack_depth is None:
            ctx = DatasetWriter(
                parent=nxdata["group"],
                name=name,
                npoints=npoints,
                attrs=attrs,
            )
            nxdata["signals"].append(name)
            writer = self._ctx_stack.enter_context(ctx)
            return _NexusDataHandler(group=group, name=name, writer=writer)
        else:
            ctx = StackDatasetWriter(
                parent=nxdata["group"],
                name=name,
                npoints=npoints,
                nstack=stack_depth,
                attrs=attrs,
            )
            nxdata["signals"].append(name)
            writer = self._ctx_stack.enter_context(ctx)
            return _NexusStackDataHandler(group=group, name=name, writer=writer)

    def _finalize(self):
        for nxdata in self._nxdata_groups.values():
            group = nxdata["group"]
            if "signal" in group.attrs:
                continue
            signals = nxdata["signals"]
            group.attrs["signal"] = signals[0]
            if len(signals) > 1:
                group.attrs["auxiliary_signals"] = signals[1:]



from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union

import numpy


class NotAvailable:
    """Use when the selected fit method does not provide a certain results."""

    pass


UNITS = {
    "Time": "s",
    "Flux": "Hz",
    "DetectorDistance": "cm",
    "DetectorArea": "cm^2",
    "SolidAngle": "sr",
    "energies": "keV",
    "raw_energy": "keV",
}


class QuantificationParameters(NamedTuple):
    I0: float  # incident photon count
    Time: float  # sec
    Flux: float  # I0/Time (ph/sec)
    DetectorDistance: float  # cm
    DetectorArea: float  # cm^2
    SolidAngle: float  # steradian
    ReferenceElement: Optional[str]
    ReferenceTransitions: Optional[str]


class XRFBatchFitResult(NamedTuple):
    # Nparameters: number of fit parameters
    # Narea_parameters: number of peak area parameters
    # Nlayers: number of layers in a multilayer matrix
    # Nspectra: number of XRF spectra
    # Nchan: original number of MCA channels
    # Nfit: fit number of MCA channels

    parameters: numpy.ndarray  # (Nparameters, Nspectra)
    uncertainties: numpy.ndarray  # (Nparameters, Nspectra)
    derivatives: Optional[numpy.ndarray]  # (Nparameters, Nchan)
    parameter_names: List[str]  # (Nparameters,)
    channels: numpy.ndarray  # (Nfit,)
    energies: numpy.ndarray  # (Nfit,)
    raw_channels: numpy.ndarray  # (Nchan,)
    raw_energies: numpy.ndarray  # (Nchan,)
    spectra: Union[numpy.ndarray, None, NotAvailable]  # (Nspectra, Nchan)
    model: Union[numpy.ndarray, None, NotAvailable]  # (Nspectra, Nchan)
    residuals: Union[numpy.ndarray, None, NotAvailable]  # (Nspectra, Nchan)
    mass_fractions: Optional[
        numpy.ndarray
    ]  # (Narea_parameters, Nspectra); None when no quantification
    mass_fractions_per_layer: Optional[
        List[numpy.ndarray]
    ]  # (Nlayers, Narea_parameters, Nspectra); None when no quantification
    mass_fraction_names: Optional[
        List[str]
    ]  # (Narea_parameters,); None when no quantification
    quantification_parameters: Union[QuantificationParameters, NotAvailable]

    @property
    def has_diagnostics(self) -> bool:
        return all(self.has_data(value) for value in ["spectra", "model", "residuals"])

    @property
    def num_spectra(self) -> int:
        if self.parameters.ndim == 2:
            return self.parameters.shape[1]
        return 0

    def has_data(self, name) -> bool:
        value = getattr(self, name)
        return value is not None and not isinstance(value, NotAvailable)



from typing import Dict

import h5py
import numpy
import pytest

from ...io.hdf5 import split_h5uri
from ...xrffit import batch
from ..utils import generate_data


@pytest.mark.parametrize("nscans", [1, 2], ids=["1_scan", "2_scans"])
@pytest.mark.parametrize("ndetectors", [1, 2], ids=["single_det", "multi_det"])
@pytest.mark.parametrize(
    "npoints_per_scan", [1, 7, 200], ids=["1_spectrum", "7_spectra", "200_spectra"]
)
@pytest.mark.parametrize("fast_fitting", [True, False], ids=["fast", "slow"])
@pytest.mark.parametrize("linear", [True, False], ids=["linear", "nonlinear"])
@pytest.mark.parametrize("samefile", [True, False], ids=["same_file", "different_file"])
def test_fit_xrf_urls(
    tmp_path, nscans, ndetectors, npoints_per_scan, fast_fitting, linear, samefile
):
    if not fast_fitting and npoints_per_scan > 10:
        pytest.skip("too slow, no extra value in testing")
    if not samefile and nscans == 1:
        pytest.skip("no extra value in testing")

    diagnostics = True
    quantification = True
    energy = 7.5
    energy_multiplier = 10

    # Generate data
    xrf_spectra_uris, spectra, parameters, configuration = generate_data(
        npoints_per_scan,
        energy,
        tmp_path=tmp_path,
        samefile=samefile,
        nscans=nscans,
        ndetectors=ndetectors,
    )

    # Raw data URI's
    detector_uris = {}
    for lst in xrf_spectra_uris:
        for xrf_spectra_uri in lst:
            file_path, data_path = xrf_spectra_uri.split("::")
            data_parts = data_path.split("/")
            bliss_scan_uri = f"{file_path}::/{'/'.join(data_parts[:2])}"
            detector_name = data_parts[-1]

            lst = detector_uris.setdefault(detector_name, [])
            lst.append(bliss_scan_uri)

    fit_io_uris = []
    output_file = tmp_path / "output.h5"
    output_root_uri = f"{output_file}::/1.1"
    for detector_name, bliss_scan_uris in detector_uris.items():
        fit_io_uri = batch.FitUris(
            bliss_scan_uris=bliss_scan_uris,
            detector_name=detector_name,
            energy_name="energy",
            output_root_uri=output_root_uri,
            xrf_spectra_uri_template="measurement/{}",
            energy_uri_template="instrument/positioners_start/{}",
        )
        fit_io_uris.append(fit_io_uri)

    # Configuration
    configuration.fit.linearfitflag = int(linear)

    # Fit
    batch.fit_xrf_urls(
        fit_io_uris,
        configuration,
        diagnostics=diagnostics,
        quantification=quantification,
        energy_multiplier=energy_multiplier,
        fast_fitting=fast_fitting,
    )

    # Validate
    for detector_name, detector_spectra in zip(detector_uris, spectra):
        xrf_results_uri = f"{output_root_uri}/fit/{detector_name}/results"
        _validate_results(
            xrf_results_uri, fast_fitting, linear, parameters, detector_spectra
        )


def _validate_results(
    xrf_results_uri: str,
    fast_fitting: bool,
    linear: bool,
    parameters: Dict[str, numpy.ndarray],
    spectra: numpy.ndarray,
):
    output_file, output_h5path = split_h5uri(xrf_results_uri)

    with h5py.File(output_file, mode="r") as h5file:
        result_group = h5file[output_h5path]
        npeakareas = 12
        nnonlinear = 12
        if fast_fitting:
            nparams = npeakareas
        else:
            nparams = npeakareas + nnonlinear

        expected = {"parameters", "uncertainties", "mass_fractions", "fit"}
        if fast_fitting:
            expected.add("derivatives")
        assert set(result_group) == expected

        # Fit results
        assert len(result_group["parameters"]) == nparams
        assert len(result_group["uncertainties"]) == nparams
        for name, values in parameters.items():
            _check_param_dataset(values, name, result_group)

        if fast_fitting:
            assert len(result_group["derivatives"]) == npeakareas + 1

        # Fit
        expected = {"data", "model", "residuals", "energy"}
        assert set(result_group["fit"]) == expected

        spectra2 = result_group["fit/data"][()]
        numpy.testing.assert_allclose(spectra, spectra2, atol=1e-10)
        model = result_group["fit/model"][()]

        residuals = result_group["fit/residuals"][()]
        residuals2 = spectra - model
        mask = ~numpy.isnan(model)
        if not fast_fitting:
            residuals2 = -residuals2
        numpy.testing.assert_allclose(residuals[mask], residuals2[mask], atol=1e-4)


def _check_param_dataset(expected_counts, dset_name, result_group):
    fit_counts = result_group[f"parameters/{dset_name}"][()]
    if expected_counts.size < 10:
        # TODO: does not always work. Weights are disabled but even when they are enabled, it does not work.
        fit_errors = 3 * result_group[f"uncertainties/{dset_name}"][()]
        diff = numpy.abs(fit_counts - expected_counts)
        assert (diff < fit_errors).all()
    diff = numpy.abs(numpy.diff(fit_counts) - 50)
    assert (diff < 5).all()



from ewoksfluo.gui.mesh_widget import OWMeshWidget
from ewoksfluo.tasks.example_data.tasks import MeshStackSingleDetector


class OWMeshStackSingleDetector(OWMeshWidget, ewokstaskclass=MeshStackSingleDetector):
    name = "Mesh stack (single detector)"
    description = "XRF test data of a stack of identical scans with one detector"

    def _init_control_area(self):
        super()._init_control_area(stack=True, multidetector=False, mosaic=False)



import logging
import os
import re
import subprocess
import time
import tracemalloc
from contextlib import contextmanager
from typing import Generator
from typing import Optional
from typing import Tuple

import numpy
import psutil
from numpy.typing import DTypeLike

logger = logging.getLogger(__name__)


def log_required_memory(
    action: str, shape: Tuple[int, ...], dtype: DTypeLike, multiplier: int = 1
) -> None:
    """Log the memory required for the processing of an array."""
    required_bytes = _calculate_array_memory(shape, dtype) * multiplier
    required_bytes_str = _format_mem_size(required_bytes)
    available_bytes = _get_available_memory_bytes()
    available_bytes_str = _format_mem_size(available_bytes)

    logger.log(
        logging.WARNING if required_bytes > available_bytes else logging.INFO,
        "Memory required to %s: %s (available: %s)",
        action,
        required_bytes_str,
        available_bytes_str,
    )


def array_chunk_size(
    shape: Tuple[int, ...], dtype: DTypeLike, multiplier: int = 1
) -> int:
    """Process array with a chunk size along the first dimension to fit in memory."""
    required_bytes = _calculate_array_memory(shape, dtype) * multiplier
    available_mem = _get_available_memory_bytes()

    if required_bytes <= available_mem:
        logger.info("Process array in 1 chunk.")
        return shape[0]

    chunk_fraction = available_mem / required_bytes
    chunk_size_dim0 = max(1, int(shape[0] * chunk_fraction))
    nchunks = numpy.ceil(shape[0] / chunk_size_dim0)
    logger.info("Process array in %d chunks.", nchunks)
    return chunk_size_dim0


def array_block_size(
    item_shape: Tuple[int, ...], dtype: DTypeLike, max_workers: int, multiplier: int = 1
) -> int:
    """Number of items that can be processing in parallel while staying within the memory limits."""
    required_bytes = _calculate_array_memory(item_shape, dtype) * multiplier
    if required_bytes == 0:
        return 0
    mem = _get_available_memory_bytes()
    worker_mem = mem / (max_workers or 1)
    return int(worker_mem / required_bytes)


def job_parameters(
    narrays: int,
    shape: Tuple[int, ...],
    dtype: DTypeLike,
    multiplier: int = 1,
    chunking_possible: bool = True,
) -> Tuple[int, int]:
    """
    Return maximum number of parallel jobs and chunk size along the first
    dimension so that CPU and memory usage stay within available limits.
    """
    if narrays <= 0 or not shape or shape[0] == 0:
        return 0, 0  # scalar or empty or no array

    # Number of arrays we can process in parallel capped by available CPUs.
    cpu_cap = get_available_cpus(exclude_current=True)
    narrays_in_parallel = min(cpu_cap, narrays)

    # Total memory required for processing full array's in parallel.
    mem_per_array = _calculate_array_memory(shape, dtype) * multiplier
    required_mem = narrays_in_parallel * mem_per_array

    # Allow when the available memory is not exceeded.
    available_mem = _get_available_memory_bytes()
    available_mem_str = _format_mem_size(available_mem)
    if required_mem <= available_mem:
        used_mem_str = _format_mem_size(required_mem)
        logger.info(
            "Process %d arrays with %d workers. Requires %s, %s available.",
            narrays,
            narrays_in_parallel,
            used_mem_str,
            available_mem_str,
        )
        return narrays_in_parallel, shape[0]

    if not chunking_possible:
        # Memory is exceeded and chunking is NOT allowed.
        # So reduce number of parallel jobs but require at
        # least one which could cause memory overuse.
        narrays_in_parallel = max(available_mem // mem_per_array, 1)
        required_mem = narrays_in_parallel * mem_per_array
        used_mem_str = _format_mem_size(required_mem)
        logger.info(
            "Process %d arrays with %d workers. Requires %s, %s available.",
            narrays,
            narrays_in_parallel,
            used_mem_str,
            available_mem_str,
        )
        return narrays_in_parallel, shape[0]

    # Memory is exceeded and chunking is allowed.
    # Reduce chunk size to fit memory but requires at
    # least chunk size 1 which could cause memory overuse.
    chunk_fraction = available_mem / required_mem
    chunk_size_dim0 = max(1, int(shape[0] * chunk_fraction))
    required_mem = narrays_in_parallel * int(mem_per_array * chunk_size_dim0 / shape[0])
    used_mem_str = _format_mem_size(required_mem)
    chunks_per_array = numpy.ceil(chunk_size_dim0 / shape[0]).astype(int)
    narray_chunks = narrays * chunks_per_array
    logger.info(
        "Process %d array chunks with %d workers. Requires %s, %s available.",
        narray_chunks,
        narrays_in_parallel,
        used_mem_str,
        available_mem_str,
    )
    return narrays_in_parallel, chunk_size_dim0


@contextmanager
def peak_memory(array: numpy.ndarray) -> Generator[None, None, None]:
    tracemalloc.start()
    try:
        yield
    finally:
        _, peak = tracemalloc.get_traced_memory()
        tracemalloc.stop()

        array_bytes = array.nbytes
        print(
            f"Array size: {array_bytes / 1e6:.2f} MB | "
            f"Peak memory: {peak / 1e6:.2f} MB | "
            f"Peak/array: {peak / array_bytes:.2f}"
        )


def memory_usage(msg: str) -> None:
    process = psutil.Process(os.getpid())
    rss_start = process.memory_info().rss
    process_start_time = process.create_time()
    t_start = time.time() - process_start_time
    print(
        f"{msg or 'Memory Usage'}: time = {t_start:.2f} sec, RSS= {rss_start / 1e6:.2f} MB"
    )


def get_available_cpus(exclude_current: bool = True) -> int:
    """Get available CPUs. At least one."""
    cpus = psutil.Process().cpu_affinity() or ()
    return max(len(cpus) - exclude_current, 1)


def _get_available_memory_bytes() -> int:
    """Get available memory in bytes. Zero or higher."""
    available = _get_available_slurm_job_memory_bytes()
    if available is not None:
        return available
    available = _get_available_cgroup_memory_bytes()
    if available is not None:
        return available
    return _get_available_system_memory_bytes()


def _get_available_system_memory_bytes() -> int:
    return psutil.virtual_memory().available


def _get_available_slurm_job_memory_bytes() -> Optional[int]:
    requested = _get_slurm_requested_memory_bytes()
    current = _get_slurm_current_memory_usage_bytes()
    if current is None or requested is None:
        return
    return max(requested - current, 0)


def _get_slurm_requested_memory_bytes() -> Optional[int]:
    job_id = os.getenv("SLURM_JOB_ID")
    if not job_id:
        return None
    try:
        result = subprocess.run(
            ["sacct", "-j", job_id, "--format=ReqMem", "-P", "-n"],
            capture_output=True,
            text=True,
            check=True,
        )
        req_mem_str = result.stdout.strip()
        return _parse_mem_size(req_mem_str)
    except Exception:
        return None


def _get_slurm_current_memory_usage_bytes() -> Optional[int]:
    job_id = os.getenv("SLURM_JOB_ID")
    if not job_id:
        return None
    try:
        # Query sstat for the job's batch step memory usage
        result = subprocess.run(
            ["sstat", "-j", job_id, "--format=MaxRSS", "-P", "-n"],
            capture_output=True,
            text=True,
            check=True,
        )
        mem_str = result.stdout.strip()
        return _parse_mem_size(mem_str)
    except Exception:
        return None


def _get_available_cgroup_memory_bytes() -> Optional[int]:
    try:
        # cgroups v2
        if os.path.exists("/sys/fs/cgroup/memory.max"):
            with open("/sys/fs/cgroup/memory.max") as f:
                limit = f.read().strip()
            if limit == "max":
                return None
            limit = int(limit)
            with open("/sys/fs/cgroup/memory.current") as f:
                current = int(f.read().strip())
            return max(limit - current, 0)

        # cgroups v1
        if os.path.exists("/sys/fs/cgroup/memory/memory.limit_in_bytes"):
            with open("/sys/fs/cgroup/memory/memory.limit_in_bytes") as f:
                limit = int(f.read().strip())
            if limit > 1 << 60:
                return None  # unlimited
            with open("/sys/fs/cgroup/memory/memory.usage_in_bytes") as f:
                current = int(f.read().strip())
            return max(limit - current, 0)

    except Exception:
        return None


def _parse_mem_size(mem_str: str) -> Optional[int]:
    """
    Parse memory string from sstat (e.g. '500M', '1.2G') to an integer in bytes.
    Returns None if parsing fails.
    """
    if not mem_str:
        return None
    mem_str = mem_str.strip().upper()
    match = re.match(r"([\d\.]+)([KMGTP])?", mem_str)
    if not match:
        return None
    value, unit = match.groups()
    try:
        value = float(value)
    except ValueError:
        return None
    unit_multipliers = {
        None: 1,
        "K": 1024,
        "M": 1024**2,
        "G": 1024**3,
        "T": 1024**4,
        "P": 1024**5,
    }
    return int(value * unit_multipliers.get(unit, 1))


def _format_mem_size(num_bytes: int) -> str:
    units = ["bytes", "KB", "MB", "GB", "TB"]
    size = float(num_bytes)
    unit_index = 0
    while size >= 1024 and unit_index < len(units) - 1:
        size /= 1024
        unit_index += 1
    return f"{size:.2f} {units[unit_index]}"


def _calculate_array_memory(shape: Tuple[int, ...], dtype: DTypeLike) -> int:
    dtype = numpy.dtype(dtype)
    num_elements = numpy.prod(shape) if shape else 1
    return num_elements * dtype.itemsize


if __name__ == "__main__":
    available = _get_available_memory_bytes()
    available_slurm = _get_available_slurm_job_memory_bytes()
    available_cgroup = _get_available_cgroup_memory_bytes()
    available_system = _get_available_system_memory_bytes()

    print(
        f"Available memory: {_format_mem_size(available) if available else float('nan')}"
    )
    print(
        f"Available memory (SLURM): {_format_mem_size(available_slurm) if available_slurm else float('nan')}"
    )
    print(
        f"Available memory (cgroup): {_format_mem_size(available_cgroup) if available_cgroup else float('nan')}"
    )
    print(
        f"Available memory (system): {_format_mem_size(available_system) if available_system else float('nan')}"
    )



from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy
from scipy.interpolate import griddata

from ..io import hdf5
from ..units import unit_registry
from . import grid_utils
from . import pad
from .optimal_grid import optimal_grid_axes


class ScatterDataInterpolator:
    """Interpolates scatter data on a regular grid."""

    def __init__(
        self,
        scatter_coordinates: Sequence[numpy.ndarray],
        scatter_coordinate_names: Sequence[str],
        scatter_coordinate_units: Sequence[Optional[str]],
        method: Optional[str] = None,
        fill_value=None,
        fix_resolution: bool = True,
        fix_limits: bool = True,
        resolution: Optional[
            Union[Sequence[float], Sequence[Tuple[float, str]]]
        ] = None,
        outside_resolution_fraction: float = 2.1,
    ):
        """
        :param scatter_coordinates: independent variables with shape `(Ndim, Nscatter)`.
        :param scatter_coordinate_names: independent variable names with shape `(Ndim,)`
        :param scatter_coordinate_units: independent variable units with shape `(Ndim,)`
        :param method: interpolation method.
        :param fill_value: value used for extrapolation.
        :param fix_resolution: fix the resolution.
        :param fix_limits: fix the limits.
        :param resolution: grid spacing in each dimension.
        :param outside_resolution_fraction: a grid point is concidered to be outside the space spanned by the scatter points
                                 when the closest scatter point is farther away than a fraction of the smallest resolution
                                 across all dimensions.
        """
        if method is None:
            method = "nearest"
        self._method = method
        if fill_value is None:
            fill_value = numpy.nan
        self._fill_value = fill_value
        self._axes_names = scatter_coordinate_names

        # Units
        self._units = scatter_coordinate_units[0]

        ureg = unit_registry()

        if len(set(scatter_coordinate_units)) > 1:
            # All dimensions will have the same units (self._units)
            assert not any(units is None for units in scatter_coordinate_units)

            if self._units:
                scatter_coordinates = [
                    (v * ureg.parse_units(u)).to(self._units).magnitude
                    for v, u in zip(scatter_coordinates, scatter_coordinate_units)
                ]

        if resolution and self._units:
            # Resolution must have the same units as scatter_coordinates
            resolution_correct_units = list()
            for value in resolution:
                if isinstance(value, tuple):
                    v, u = value
                    value = (v * ureg.parse_units(u)).to(self._units).magnitude
                    resolution_correct_units.append(value)
                else:
                    resolution_correct_units.append(value)
            resolution = resolution_correct_units

        # Scatter coordinates
        assert len(scatter_coordinate_names) == len(scatter_coordinates)
        assert len(scatter_coordinate_units) == len(scatter_coordinates)

        nscatter = set()
        for i, coordi in enumerate(scatter_coordinates):
            if coordi.ndim != 1:
                raise ValueError(f"Coordinate {i} must be provided as a 1D array")
            if coordi.size == 0:
                raise ValueError(f"Coordinate {i} is empty")
            nscatter.add(len(coordi))

        self._nscatter = min(nscatter)

        if len(nscatter) > 1:
            scatter_coordinates = [
                coordi[: self._nscatter] for coordi in scatter_coordinates
            ]

        self._scatter_coordinates = numpy.vstack(
            scatter_coordinates
        ).T  # (Nscatter, Ndim)

        # Grid coordinates
        self._grid_axes = optimal_grid_axes(
            self._scatter_coordinates,
            fix_resolution=fix_resolution,
            fix_limits=fix_limits,
            resolution=resolution,
        )  # [(N0,), (N1,), ...]

        self._grid_shape = tuple(len(arr) for arr in self._grid_axes)  # (N0, N1, ...)
        self._expanded_grid_coordinates = grid_utils.expanded_grid_coordinates(
            self._grid_axes
        )  # (N0*N1*..., Ndim)

        # Difference between scatter and grid coordinates
        self._closest_scatter_index, self._closest_scatter_distance = (
            grid_utils.closest_point(
                self._expanded_grid_coordinates, self._scatter_coordinates
            )
        )
        self._closest_grid_index, self._closest_grid_distance = (
            grid_utils.closest_point(
                self._scatter_coordinates, self._expanded_grid_coordinates
            )
        )
        distance_max = min(
            abs(axis[1] - axis[0]) * outside_resolution_fraction
            for axis in self._grid_axes
        )
        self._grid_coordinates_outside = self._closest_scatter_distance > distance_max

        # self._plot_coordinates()

    @property
    def scan_size(self) -> int:
        return self._nscatter

    @property
    def grid_size(self) -> int:
        return self._expanded_grid_coordinates[0]

    @property
    def grid_shape(self) -> Tuple[int]:
        """
        :returns: tuple `(N0, N1, ...)` with length: `Ndim`
        """
        return self._grid_shape

    @property
    def grid_ndim(self) -> int:
        return len(self._grid_shape)

    @property
    def axes_names(self) -> List[str]:
        return self._axes_names

    @property
    def units(self) -> Optional[str]:
        return self._units

    @property
    def grid_axes(self) -> List[numpy.ndarray]:
        """
        :returns: `Ndim` arrays with shapes `[(N0,), (N1,), ...]`
        """
        return self._grid_axes

    @property
    def expanded_grid_coordinates(self) -> numpy.ndarray:
        """
        :returns: shape `(N0*N1*..., Ndim)`
        """
        return self._expanded_grid_coordinates

    @property
    def grid_coordinates_outside(self) -> numpy.ndarray:
        """Mask for grid coordinates with shape `(N0*N1*..., )` that are outside the scatter cloud.

        :returns: shape `(N0*N1*...,)`
        """
        return self._grid_coordinates_outside

    @property
    def scatter_coordinates(self) -> numpy.ndarray:
        """
        :returns: shape `(Nscatter, Ndim)`
        """
        return self._scatter_coordinates

    def regrid(self, scatter_data: numpy.ndarray) -> numpy.ndarray:
        """
        :param scatter_data: flat list of data values `(Nscatter, M0, M1, ...)` with
                             `(M0, M1, ...)` the shape of one detector data value
        :returns: regridded data values `(N0, N1, ..., M0, M1, ...)`
        """
        if scatter_data.ndim == 0:
            raise ValueError("Data cannot be a scalar")
        if scatter_data.size == 0:
            raise ValueError("Data is empty")

        ndata = len(scatter_data)
        if ndata > self._nscatter:
            scatter_data = scatter_data[: self._nscatter]
        elif ndata < self._nscatter:
            scatter_data = pad.pad_array(scatter_data, self._nscatter)

        if scatter_data.ndim == 1:
            return self._interpolate_0d_detector(scatter_data)

        return self._interpolate_nd_detector(scatter_data)

    def _interpolate_0d_detector(self, scatter_data: numpy.ndarray) -> numpy.ndarray:
        interpolated_data = _interpolate(
            scatter_coordinates=self._scatter_coordinates,
            scatter_data=scatter_data,
            interp_coordinates=self._expanded_grid_coordinates,
            interp_coordinates_outside=self._grid_coordinates_outside,
            method=self._method,
            fill_value=self._fill_value,
        )
        return interpolated_data.reshape(*self._grid_shape)

    def _interpolate_nd_detector(self, scatter_data: numpy.ndarray) -> numpy.ndarray:
        nscatter = scatter_data.shape[0]
        data_shape = scatter_data.shape[1:]
        scatter_data_flat_detector = scatter_data.reshape((nscatter, -1))
        nscatter, data_size = scatter_data_flat_detector.shape

        interpolated_data = numpy.array(
            [
                _interpolate(
                    scatter_coordinates=self._scatter_coordinates,
                    scatter_data=scatter_data_flat_detector[:, i],
                    interp_coordinates=self._expanded_grid_coordinates,
                    interp_coordinates_outside=self._grid_coordinates_outside,
                    method=self._method,
                    fill_value=self._fill_value,
                )
                for i in range(data_size)
            ]
        ).T

        return interpolated_data.reshape(*self._grid_shape, *data_shape)

    def _plot_coordinates(self) -> None:
        import matplotlib.pyplot as plt

        outside = self.grid_coordinates_outside
        inside = ~outside

        plt.plot(
            self.scatter_coordinates[:, 1],
            self.scatter_coordinates[:, 0],
            ".",
            color="blue",
        )
        plt.plot(
            self.expanded_grid_coordinates[inside, 1],
            self.expanded_grid_coordinates[inside, 0],
            "o",
            mfc="none",
            color="green",
        )
        plt.plot(
            self.expanded_grid_coordinates[outside, 1],
            self.expanded_grid_coordinates[outside, 0],
            "o",
            mfc="none",
            color="red",
        )
        plt.xlabel(self.axes_names[1])
        plt.ylabel(self.axes_names[0])
        plt.axis("equal")
        plt.show()

    def save_coordinates_as_nxdata(self, parent: hdf5.GroupType) -> None:
        nxcollection = parent.create_group("coordinates")
        nxcollection.attrs["NX_class"] = "NXcollection"

        nxdata = nxcollection.create_group("scatter_coordinates")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata["title"] = "Scatter coordinate indices"
        self._save_nxdata_scatter_axes(nxdata)
        signal = numpy.arange(len(self.scatter_coordinates))
        self._save_nxdata_signal(nxdata, "indices", signal, units=False)

        nxdata = nxcollection.create_group("grid_coordinates")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata["title"] = "Grid nodes that have data"
        self._save_nxdata_grid_axes(nxdata)
        signal = ~self.grid_coordinates_outside
        self._save_nxdata_signal(nxdata, "has_data", signal, units=False)

        nxdata = nxcollection.create_group("coordinates")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata["title"] = "Grid nodes (0 or 1) and Scatter coordinates (2)"
        self._save_nxdata_scatter_and_grid_axes(nxdata)
        signal = numpy.concatenate(
            (
                numpy.full((len(self.scatter_coordinates),), 2),
                ~self.grid_coordinates_outside,
            )
        )
        self._save_nxdata_signal(nxdata, "coordinate_type", signal, units=False)

        nxdata = nxcollection.create_group("closest_grid_distance")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata["title"] = "Closest grid node distance for every scatter coordinate"
        self._save_nxdata_scatter_axes(nxdata)
        signal = self._closest_grid_distance
        self._save_nxdata_signal(nxdata, "distance", signal, units=True)

        nxdata = nxcollection.create_group("closest_scatter_distance")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata["title"] = "Closest scatter coordinate distance for every grid node"
        self._save_nxdata_grid_axes(nxdata)
        signal = self._closest_scatter_distance.copy()
        signal[self._grid_coordinates_outside] = numpy.nan
        self._save_nxdata_signal(nxdata, "distance", signal, units=True)

    def _save_nxdata_signal(
        self, nxdata: hdf5.GroupType, name: str, data: numpy.ndarray, units: bool = True
    ) -> None:
        nxdata.attrs["signal"] = name
        dset = nxdata.create_dataset(name=name, data=data)
        if units:
            self._save_nxdata_units(dset, name)

    def _save_nxdata_scatter_axes(self, nxdata: hdf5.GroupType) -> None:
        nxdata.attrs["axes"] = self.axes_names[::-1]
        for name, values in zip(self.axes_names, self.scatter_coordinates.T):
            dset = nxdata.create_dataset(name=name, data=values)
            self._save_nxdata_units(dset, name)

    def _save_nxdata_grid_axes(self, nxdata: hdf5.GroupType) -> None:
        nxdata.attrs["axes"] = self.axes_names[::-1]
        for name, values in zip(self.axes_names, self.expanded_grid_coordinates.T):
            dset = nxdata.create_dataset(name=name, data=values)
            self._save_nxdata_units(dset, name)

    def _save_nxdata_scatter_and_grid_axes(self, nxdata: hdf5.GroupType) -> None:
        nxdata.attrs["axes"] = self.axes_names[::-1]
        for name, scatter_values, grid_values in zip(
            self.axes_names,
            self.scatter_coordinates.T,
            self.expanded_grid_coordinates.T,
        ):
            values = numpy.concatenate((scatter_values, grid_values))
            dset = nxdata.create_dataset(name=name, data=values)
            self._save_nxdata_units(dset, name)

    def _save_nxdata_units(self, dset: hdf5.DatasetType, name: str) -> None:
        if self.units is not None:
            dset.attrs["units"] = self.units
            dset.attrs["long_name"] = f"{name} ({self.units})"


def _interpolate(
    scatter_coordinates: numpy.ndarray,
    scatter_data: numpy.ndarray,
    interp_coordinates: numpy.ndarray,
    interp_coordinates_outside: numpy.ndarray,
    method: str,
    fill_value: float,
) -> numpy.ndarray:
    """
    :param scatter_coordinates: scatter coordinates with shape `(Nscatter, Ndim)`
    :param scatter_data: flat list of scatter data values `(Nscatter,)`
    :param interp_coordinates: interpolate coordinates with shape `(Ninterp, Ndim)`
    :param interp_coordinates_outside: boolean array with shape `(Ninterp,)`
    :param method:
    :param fill_value:
    :returns: interpolated data values `(Ninterp,)`
    """
    interp_data = griddata(
        scatter_coordinates,
        scatter_data,
        interp_coordinates,
        method=method,
        fill_value=fill_value,
    )
    if numpy.isnan(fill_value) and numpy.issubdtype(interp_data.dtype, numpy.integer):
        fill_value = 0
    interp_data[interp_coordinates_outside] = fill_value
    return interp_data



from ewoksfluo.gui.fit_widget import OWFitWidget
from ewoksfluo.tasks.fit.tasks import FitSingleScanSingleDetector


class OWFitSingleScanSingleDetector(
    OWFitWidget, ewokstaskclass=FitSingleScanSingleDetector
):
    name = "Fit scan (single detector)"
    description = "Fit one scan with one detector"

    def _init_control_area(self):
        super()._init_control_area(stack=False, multidetector=False)



from typing import NamedTuple

import numpy


class ScanData(NamedTuple):
    group: str
    name: str
    detector_type: str
    data: numpy.ndarray  # npoints x ...
    local_alias: str
    global_alias: str


class ZapInfo(NamedTuple):
    """SPEC zap scan information"""

    directory: str
    radix: str
    scannb: int
    zapnb: int



import sysconfig

WIDGET_HELP_PATH = (
    # Development documentation (make htmlhelp in ./doc)
    ("{DEVELOP_ROOT}/doc/_build/htmlhelp/index.html", None),
    # Documentation included in wheel
    ("{}/help/ewoksfluo/index.html".format(sysconfig.get_path("data")), None),
    # Online documentation url
    ("https://ewoksfluo.readthedocs.io", ""),
)


# Entry point for main Orange categories/widgets discovery
def widget_discovery(discovery):
    from ewoksorange.pkg_meta import get_distribution

    dist = get_distribution("ewoksfluo")
    pkgs = [
        "orangecontrib.ewoksfluo.categories.input",
        "orangecontrib.ewoksfluo.categories.fit",
        "orangecontrib.ewoksfluo.categories.sum_detectors",
        "orangecontrib.ewoksfluo.categories.normalization",
        "orangecontrib.ewoksfluo.categories.regrid",
        "orangecontrib.ewoksfluo.categories.raw_counters",
        "orangecontrib.ewoksfluo.categories.demo",
    ]
    for pkg in pkgs:
        discovery.process_category_package(pkg, distribution=dist)



from typing import Dict

import h5py
import numpy
import pytest

from ....io.hdf5 import split_h5uri
from ....xrffit.legacy import outputbuffer_context
from ....xrffit.legacy import perform_batch_fit
from ....xrffit.pymca_config import pymca_configdict_from_model
from ....xrffit.pymca_config import pymca_configdict_to_file
from ...utils import generate_data


@pytest.mark.parametrize("nscans", [1, 2], ids=["1_scan", "2_scans"])
@pytest.mark.parametrize(
    "npoints_per_scan", [1, 7, 200], ids=["1_spectrum", "7_spectra", "200_spectra"]
)
@pytest.mark.parametrize("fast_fitting", [True, False], ids=["fast", "slow"])
@pytest.mark.parametrize("linear", [True, False], ids=["linear", "nonlinear"])
@pytest.mark.parametrize("output_handler", ["nexus", "pymca"])
@pytest.mark.parametrize("samefile", [True, False], ids=["same_file", "different_file"])
def test_perform_batch_fit(
    tmp_path, nscans, npoints_per_scan, fast_fitting, linear, output_handler, samefile
):
    if not fast_fitting and npoints_per_scan > 10:
        pytest.skip("too slow, no extra value in testing")
    if not samefile and nscans == 1:
        pytest.skip("no extra value in testing")

    diagnostics = True
    figuresofmerit = True
    quantification = True
    energy = 7.5
    energy_multiplier = 10

    # Generate data
    xrf_spectra_uris, spectra, parameters, config = generate_data(
        npoints_per_scan,
        energy,
        tmp_path=tmp_path,
        samefile=samefile,
        nscans=nscans,
        ndetectors=1,
    )
    # Single detector:
    xrf_spectra_uris = xrf_spectra_uris[0]
    spectra = spectra[0]

    # Output
    output_root_uri = str(tmp_path / "output.h5::/1.1/fit")
    xrf_results_uri = str(tmp_path / "output.h5::/1.1/fit/results")

    # Configuration
    config.fit.linearfitflag = int(linear)
    config_filename = str(tmp_path / "config.cfg")
    pymca_configdict_to_file(pymca_configdict_from_model(config), config_filename)

    # Perform fit
    with outputbuffer_context(
        output_root_uri,
        diagnostics=diagnostics,
        figuresofmerit=figuresofmerit,
        output_handler=output_handler,
    ) as output_buffer:
        kwargs = {
            "xrf_spectra_uris": xrf_spectra_uris,
            "pymca_configuration": config_filename,
            "output_buffer": output_buffer,
            "energy": energy,
            "energy_multiplier": energy_multiplier,
            "fast_fitting": fast_fitting,
            "quantification": quantification,
        }

        if not samefile and nscans > 1:
            with pytest.raises(
                ValueError, match="cannot handle scans in different files"
            ):
                perform_batch_fit(**kwargs)
            return

        perform_batch_fit(**kwargs)

        assert output_buffer.xrf_results_uri == xrf_results_uri

    if output_handler == "pymca":
        # The pymca output handler keeps the data dimensions
        if nscans > 1 and npoints_per_scan > 1:
            newshape = (nscans, npoints_per_scan, spectra.shape[-1])
            spectra = spectra.reshape(newshape)
        else:
            spectra = spectra[None, ...]

    _validate_results(
        xrf_results_uri, output_handler, fast_fitting, linear, parameters, spectra
    )


def _validate_results(
    xrf_results_uri: str,
    output_handler: str,
    fast_fitting: bool,
    linear: bool,
    parameters: Dict[str, numpy.ndarray],
    spectra: numpy.ndarray,
):
    output_file, output_h5path = split_h5uri(xrf_results_uri)
    # Validate results
    with h5py.File(output_file, mode="r") as h5file:
        result_group = h5file[output_h5path]
        nparams = 12
        nobservations = 1021

        # Fit results
        if output_handler == "pymca":
            # includes *_error softlinks
            assert len(result_group["parameters"]) == 2 * nparams
        else:
            assert len(result_group["parameters"]) == nparams
        assert len(result_group["uncertainties"]) == nparams
        for name, values in parameters.items():
            _check_param_dataset(values, name, result_group)

        # Diagnostics
        if fast_fitting:
            assert set(result_group["diagnostics"]) == {
                "nFreeParameters",
                "nObservations",
            }
        else:
            assert set(result_group["diagnostics"]) == {
                "chisq",
                "nFreeParameters",
                "nObservations",
            }
        numpy.testing.assert_array_equal(
            result_group["diagnostics/nFreeParameters"][()], nparams
        )
        numpy.testing.assert_array_equal(
            result_group["diagnostics/nObservations"][()], nobservations
        )

        if fast_fitting:
            if output_handler == "pymca":
                # + channels and energy
                assert len(result_group["derivatives"]) == nparams + 2
            else:
                # + energy
                assert len(result_group["derivatives"]) == nparams + 1

        # Fit
        if output_handler == "pymca":
            if not fast_fitting and not linear:
                expected = {
                    "data",
                    "model",
                    "residuals",
                    "dim2",
                    "dim0",
                    "dim1",
                }
            else:
                expected = {
                    "data",
                    "model",
                    "residuals",
                    "energy",
                    "channels",
                    "dim0",
                    "dim1",
                }
        else:
            if not fast_fitting and not linear:
                expected = {"data", "model", "residuals", "dim2"}
            else:
                expected = {"data", "model", "residuals", "energy"}
        assert set(result_group["fit"]) == expected

        spectra2 = result_group["fit/data"][()]
        numpy.testing.assert_allclose(spectra, spectra2, atol=1e-10)
        model = result_group["fit/model"][()]

        residuals = result_group["fit/residuals"][()]
        residuals2 = spectra - model
        mask = ~numpy.isnan(model)
        if not fast_fitting:
            residuals2 = -residuals2
        numpy.testing.assert_allclose(residuals[mask], residuals2[mask], atol=1e-4)


def _check_param_dataset(expected_counts, dset_name, result_group):
    fit_counts = result_group[f"parameters/{dset_name}"][()]
    if expected_counts.size < 10:
        # TODO: does not always work. Weights are disabled but even when they are enabled, it does not work.
        fit_errors = 3 * result_group[f"uncertainties/{dset_name}"][()]
        diff = numpy.abs(fit_counts - expected_counts)
        assert (diff < fit_errors).all()
    diff = numpy.abs(numpy.diff(fit_counts) - 50)
    assert (diff < 5).all()



from ewoksfluo.gui.mesh_widget import OWMeshWidget
from ewoksfluo.tasks.example_data.tasks import MeshSingleScanMultiDetector


class OWMeshSingleScanMultiDetector(
    OWMeshWidget,
    ewokstaskclass=MeshSingleScanMultiDetector,
):
    name = "Mesh (multi detector)"
    description = "XRF test data of one scan with multiple detectors"

    def _init_control_area(self):
        super()._init_control_area(stack=False, multidetector=True, mosaic=False)



import logging
from typing import Dict
from typing import List

from PyMca5.PyMcaPhysics.xrf import Elements
from PyMca5.PyMcaPhysics.xrf import KShell
from PyMca5.PyMcaPhysics.xrf import LShell
from PyMca5.PyMcaPhysics.xrf import MShell

from ..model import PyMcaXrfConfiguration

logger = logging.getLogger(__name__)


def split_mlines(configuration: PyMcaXrfConfiguration, mlines: Dict[str, List[str]]):
    """
    :param configuration:
    :param dict mlines: for example `{"Pb":["M4", "M5]}}`
    """
    split = False
    for element, new_lines in mlines.items():
        if element in configuration.peaks:
            old_lines = configuration.peaks[element]
            if "M" in old_lines:
                lines = [group for group in old_lines if group != "M"]
                configuration.peaks[element] = lines + new_lines
                logger.warning(
                    "PyMca configuration: modified peak groups of %s to %s",
                    element,
                    configuration.peaks[element],
                )
                split = True

    if split:
        _patch_mlines()
    else:
        _unpatch_mlines()


def _patch_mlines():
    global _ORIGINAL_NAMES, _ORIGINAL_TRANSITIONS, _ORIGINAL_RATES
    if Elements.ElementXrays == _PATCHED_NAMES:
        return
    if _ORIGINAL_NAMES is None:
        _ORIGINAL_NAMES = Elements.ElementXray
        _ORIGINAL_TRANSITIONS = Elements.ElementShellTransitions
        _ORIGINAL_RATES = Elements.ElementShellRates

    Elements.ElementXrays = _PATCHED_NAMES
    Elements.ElementShellTransitions = _PATCHED_TRANSITIONS
    Elements.ElementShellRates = _PATCHED_RATES


def _unpatch_mlines():
    if Elements.ElementXrays == _ORIGINAL_NAMES or _ORIGINAL_NAMES is None:
        return
    Elements.ElementXrays = _ORIGINAL_NAMES
    Elements.ElementShellTransitions = _ORIGINAL_TRANSITIONS
    Elements.ElementShellRates = _ORIGINAL_RATES


_ORIGINAL_NAMES = None
_ORIGINAL_TRANSITIONS = None
_ORIGINAL_RATES = None


_PATCHED_NAMES = [
    "K xrays",
    "Ka xrays",
    "Kb xrays",
    "L xrays",
    "L1 xrays",
    "L2 xrays",
    "L3 xrays",
    "M xrays",
    "M1 xrays",
    "M2 xrays",
    "M3 xrays",
    "M4 xrays",
    "M5 xrays",
]
_PATCHED_TRANSITIONS = [
    KShell.ElementKShellTransitions,
    KShell.ElementKAlphaTransitions,
    KShell.ElementKBetaTransitions,
    LShell.ElementLShellTransitions,
    LShell.ElementL1ShellTransitions,
    LShell.ElementL2ShellTransitions,
    LShell.ElementL3ShellTransitions,
    [s + "*" for s in MShell.ElementMShellTransitions],
    MShell.ElementM1ShellTransitions,
    MShell.ElementM2ShellTransitions,
    MShell.ElementM3ShellTransitions,
    MShell.ElementM4ShellTransitions,
    MShell.ElementM5ShellTransitions,
]
_PATCHED_RATES = [
    KShell.ElementKShellRates,
    KShell.ElementKAlphaRates,
    KShell.ElementKBetaRates,
    LShell.ElementLShellRates,
    LShell.ElementL1ShellRates,
    LShell.ElementL2ShellRates,
    LShell.ElementL3ShellRates,
    MShell.ElementMShellRates,
    MShell.ElementM1ShellRates,
    MShell.ElementM2ShellRates,
    MShell.ElementM3ShellRates,
    MShell.ElementM4ShellRates,
    MShell.ElementM5ShellRates,
]



from typing import Optional

import numpy
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import ClassMcaTheory
from PyMca5.PyMcaPhysics.xrf.ConcentrationsTool import ConcentrationsTool

from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import pymca_configdict_from_model
from ..pymca_config import pymca_configdict_to_model
from ._utils.calibration import channel_info
from ._utils.calibration import energy_info
from ._utils.limits import expand_data
from .types import NotAvailable
from .types import XRFBatchFitResult


def fit_xrf_spectra(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Slow fitting refers to fitting XRF spectra one-by-one.
    This function does not use the high-level API from PyMca.
    It uses the low-level API to

    - calculate elemental spectra
    - calculate mass fractions from fit results
    - strip spectral background
    - and perform linear least-squares fit.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    if not positive_peak_areas:
        raise NotImplementedError("'positive_peak_areas' must be True")
    if live_times is not None:
        raise NotImplementedError("'live_times' is not supported")
    if not individual_weights:
        raise NotImplementedError("'individual_weights' must be True")

    mca_theory = ClassMcaTheory()
    configuration = pymca_configdict_to_model(
        mca_theory.configure(pymca_configdict_from_model(configuration))
    )

    parameters = []
    uncertainties = []
    model = []
    residuals = []
    parameter_names = mca_theory.PARAMETERS
    if quantification:
        mass_fractions = []
        mass_fraction_names = None
        mass_fractions_per_layer = None

        ctool = ConcentrationsTool()
        ctool_config = ctool.configure()
        ctool_config.update(configuration.concentrations.model_dump(mode="python"))
        for spectrum in xrf_spectra:
            mca_theory.setData(y=spectrum)
            mca_theory.estimate()
            fitresult, result = mca_theory.startfit(digest=1)

            fittedpar, chisq, sigmapar, niter, lastdeltachi = fitresult
            parameters.append(fittedpar)
            uncertainties.append(sigmapar)
            if diagnostics:
                model.append(result["yfit"])
                residuals.append(result["yfit"] - result["ydata"])

            digest = {}
            digest["fitresult"] = fitresult
            digest["result"] = result
            digest["result"]["config"] = mca_theory.configure()  # deepcopy
            wresult = ctool.processFitResult(
                config=ctool_config,
                fitresult=digest,
                elementsfrommatrix=False,
                fluorates=mca_theory._fluoRates,
            )

            if mass_fraction_names is None:
                nlayers = len(wresult["layerlist"])
                mass_fraction_names = wresult["groups"]
                mass_fractions_per_layer = [[] for _ in range(nlayers)]

            mass_fractions.append(
                [wresult["mass fraction"][name] for name in mass_fraction_names]
            )
            for layer, lst_dest in zip(wresult["layerlist"], mass_fractions_per_layer):
                lst_dest.append(
                    [
                        wresult[layer]["mass fraction"][name]
                        for name in mass_fraction_names
                    ]
                )

        mass_fractions = numpy.asarray(mass_fractions).T
        mass_fractions_per_layer = [
            numpy.asarray(mf_layer).T for mf_layer in mass_fractions_per_layer
        ]
    else:
        mass_fractions = None
        mass_fractions_per_layer = None
        mass_fraction_names = None

        for spectrum in xrf_spectra:
            mca_theory.setData(y=spectrum)
            mca_theory.estimate()
            fitresult, result = mca_theory.startfit(digest=1)

            fittedpar, chisq, sigmapar, niter, lastdeltachi = fitresult
            parameters.append(fittedpar)
            uncertainties.append(sigmapar)

            if diagnostics:
                model.append(result["yfit"])
                residuals.append(result["yfit"] - result["ydata"])

    parameters = numpy.stack(parameters, axis=1)
    uncertainties = numpy.stack(uncertainties, axis=1)

    channels, raw_channels = channel_info(mca_theory)
    num_fit_observations = len(channels)
    num_raw_observations = len(raw_channels)

    if diagnostics:
        spectra = xrf_spectra
        model = numpy.array(model)
        residuals = numpy.array(residuals)
        if num_fit_observations != num_raw_observations:
            # Expand with NaN's so the model, residuals and base profiles match to
            # orginal number of observations.
            model = expand_data(channels, model, num_raw_observations, axis=1)
            residuals = expand_data(channels, residuals, num_raw_observations, axis=1)
    else:
        spectra = None
        model = None
        residuals = None

    energies = energy_info(mca_theory, channels)
    raw_energies = energy_info(mca_theory, raw_channels)

    return XRFBatchFitResult(
        parameters=parameters,
        uncertainties=uncertainties,
        derivatives=None,
        parameter_names=parameter_names,
        channels=channels,
        energies=energies,
        raw_channels=raw_channels,
        raw_energies=raw_energies,
        spectra=spectra,
        model=model,
        residuals=residuals,
        mass_fractions=mass_fractions,
        mass_fractions_per_layer=mass_fractions_per_layer,
        mass_fraction_names=mass_fraction_names,
        quantification_parameters=NotAvailable(),
    )



from typing import Generator
from typing import Tuple

import numpy
import psutil

from ..math import grid_utils


def test_closest_point():
    coordinates1 = numpy.random.rand(1000, 2)
    coordinates2 = numpy.random.rand(1001, 2)

    indices1, distances1 = grid_utils.closest_point(coordinates1, coordinates2)
    indices2, distances2 = _closest_point_brute_force(coordinates1, coordinates2)

    assert len(indices1) == len(coordinates1)
    assert len(indices2) == len(coordinates1)
    assert len(distances1) == len(coordinates1)
    assert len(distances2) == len(coordinates1)

    assert (indices1 == indices2).all()
    assert (distances1 == distances2).all()


def _closest_point_brute_force(
    coordinates1: numpy.ndarray,
    coordinates2: numpy.ndarray,
    chunk_memory_fraction: float = 0.10,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Find the closest point in `coordinates2` for each point in `coordinates1`.

    :param coordinates1: Coordinates with shape `(N1, Ndim)`
    :param coordinates2: Coordinates with shape `(N2, Ndim)`
    :param chunk_memory_fraction: Fraction of available memory to use for each batch
    :returns: A tuple containing:
            - indices of the closest points in `coordinates2` for each point in `coordinates1` with shape `(N1,)`
            - distances to the closest points with shape `(N1,)`
    """
    N1, Ndim1 = coordinates1.shape
    _, Ndim2 = coordinates2.shape
    assert Ndim1 == Ndim2

    closest_indices = numpy.empty(N1, dtype=int)
    closest_distances = numpy.empty(N1, dtype=float)

    for start, stop, differences in _iter_coordinate_differences(
        coordinates1, coordinates2, chunk_memory_fraction
    ):
        # differences: (Nchunk1i, N2, Ndim)
        distances = numpy.linalg.norm(differences, axis=2)  # (Nchunk1i, N2)
        chunk_indices = numpy.argmin(distances, axis=1)  # (Nchunk1i,)
        closest_indices[start:stop] = chunk_indices
        closest_distances[start:stop] = distances[
            list(range(len(chunk_indices))), chunk_indices
        ]

    return closest_indices, closest_distances


def _iter_coordinate_differences(
    coordinates1: numpy.ndarray,
    coordinates2: numpy.ndarray,
    chunk_memory_fraction: float = 0.10,
) -> Generator[Tuple[int, int, numpy.ndarray], None, None]:
    """
    Iterator over the difference between two sets of coordinates in chunks of the second set.

    :param coordinates1: coordinates with shape `(N1, Ndim)`
    :param coordinates2: coordinates with shape `(N2, Ndim)`
    :param chunk_memory_fraction: yielded chunks take a fraction of the currently
                                available memory at the moment of yielding
    :yields: difference between coordinates with shape `(Nchunk1i, N2, Ndim)`
            which totals to `(N1, N2, Ndim)` when concatenating all chunks
    """
    N1, Ndim1 = coordinates1.shape
    _, Ndim2 = coordinates2.shape
    assert Ndim1 == Ndim2

    coordinates2_size = coordinates2.size * coordinates2.itemsize
    coordinates2 = coordinates2[numpy.newaxis, ...]  # (1, N2, Ndim)

    start = 0
    while start < N1:
        chunk_memory = int(psutil.virtual_memory().available * chunk_memory_fraction)
        Nchunk1i = max(1, int(chunk_memory / coordinates2_size))

        stop = min(start + Nchunk1i, N1)
        chunk1 = coordinates1[start:stop, numpy.newaxis, :]  # (Nchunk1i, 1, Ndim)

        differences = numpy.abs(coordinates2 - chunk1)  # (Nchunk1i, N2, Ndim)

        yield start, stop, differences

        start += Nchunk1i



import datetime
import logging
import threading
import time

logger = logging.getLogger(__name__)


class ProgressLogger:
    def __init__(self, interval: float = 5.0):
        """
        :param interval: seconds between logging updates
        """
        self._interval = interval
        self._seconds_read = 0.0
        self._seconds_process = 0.0
        self._seconds_write = 0.0
        self._processed_num_spectra = 0
        self._expected_num_spectra = 0

        self._lock = threading.Lock()
        self._stop_event = threading.Event()
        self._thread = None
        self._t0 = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def start(self):
        """Start background logging."""
        self._stop_event.clear()
        if self._thread is None or not self._thread.is_alive():
            self._thread = threading.Thread(target=self._run, daemon=True)
            self._thread.start()

    def stop(self):
        """Stop background logging."""
        self._stop_event.set()
        if self._thread is not None:
            self._thread.join()

    def _run(self):
        """Background thread method."""
        self._t0 = time.perf_counter()
        while not self._stop_event.is_set():
            time.sleep(self._interval)
            self.log_progress()

    def log_progress(self):
        """Log current progress."""
        with self._lock:
            self._log_progress()

    def _log_progress(self):
        total_time = self._seconds_read + self._seconds_process + self._seconds_write
        pct_read = (self._seconds_read / total_time * 100) if total_time else 0
        pct_process = (self._seconds_process / total_time * 100) if total_time else 0
        pct_write = (self._seconds_write / total_time * 100) if total_time else 0

        # Compute milliseconds per spectrum
        if self._processed_num_spectra > 0:
            rate_total = (total_time / self._processed_num_spectra) * 1000
            rate_read = (self._seconds_read / self._processed_num_spectra) * 1000
            rate_process = (self._seconds_process / self._processed_num_spectra) * 1000
            rate_write = (self._seconds_write / self._processed_num_spectra) * 1000
        else:
            rate_total = rate_read = rate_process = rate_write = 0.0

        # Elapsed time
        elapsed_time = time.perf_counter() - self._t0
        elapsed_td = datetime.timedelta(seconds=int(elapsed_time))
        elapsed_str = str(elapsed_td)

        # ETA string
        eta_str = ""
        progress = 0.0
        if self._expected_num_spectra and self._processed_num_spectra:
            progress = self._processed_num_spectra / self._expected_num_spectra * 100
            remaining = max(self._expected_num_spectra - self._processed_num_spectra, 0)
            eta_seconds = (elapsed_time / self._processed_num_spectra) * remaining
            eta_td = datetime.timedelta(seconds=int(eta_seconds))
            eta_str = f" | ETA: {str(eta_td)}"

        # Dynamically derive field width from expected_num_spectra
        width = (
            len(str(self._expected_num_spectra)) if self._expected_num_spectra else 1
        )

        logger.info(
            f"%{width}s/%{width}s %4.1f %% | %.1f ms/sp | r/p/w %.1f/%.1f/%.1f ms/sp %4.1f/%4.1f/%4.1f %% | Elapsed: %s%s",
            int(self._processed_num_spectra),
            int(self._expected_num_spectra),
            progress,
            rate_total,
            rate_read,
            rate_process,
            rate_write,
            pct_read,
            pct_process,
            pct_write,
            elapsed_str,
            eta_str,
        )

    def accumulate(
        self,
        read=0.0,
        process=0.0,
        write=0.0,
        processed_num_spectra=0,
        expected_num_spectra=0,
    ):
        """Accumulate statistics."""
        with self._lock:
            self._seconds_read += read
            self._seconds_process += process
            self._seconds_write += write
            self._processed_num_spectra += processed_num_spectra
            self._expected_num_spectra += expected_num_spectra



from typing import List
from typing import NamedTuple
from typing import Optional

import numpy
from PyMca5.PyMcaMath.fitting import Gefit
from PyMca5.PyMcaMath.fitting import SpecfitFuns
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import ClassMcaTheory

from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import pymca_configdict_from_model
from ..pymca_config import pymca_configdict_to_model
from ._utils.calibration import channel_info
from ._utils.calibration import energy_info
from ._utils.concentrations import calculate_massfractions
from ._utils.limits import expand_data
from ._utils.linear_fit import fit_profiles
from .types import XRFBatchFitResult


def fit_xrf_spectra(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Fast fitting refers to linear fitting of several spectra by solving a single
    system of equations. This function does not use the high-level API from PyMca.
    It uses the low-level API to

    - calculate elemental spectra
    - calculate mass fractions from fit results
    - strip spectral background
    - and perform linear least-squares fit.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    if xrf_spectra.ndim != 2:
        raise ValueError("XRF spectra must be a 2D array (num_spectra x num_channels)")
    if live_times is not None:
        if len(live_times) != len(xrf_spectra):
            raise ValueError(
                "Number live time values is not equal to the number of XRF spectra"
            )

    if diagnostics:
        spectra = xrf_spectra
    else:
        spectra = None

    # Effective live time normalization
    useautotime = bool(configuration.concentrations.useautotime)
    usematrix = bool(configuration.concentrations.usematrix)
    autotime = quantification and useautotime and not usematrix
    if autotime:
        norm_factors = configuration.concentrations.time / live_times
    else:
        norm_factors = 1
    configuration.concentrations.useautotime = 0

    # Initialize McaTheory instance
    mca_theory = ClassMcaTheory()
    mca_theory.enableOptimizedLinearFit()
    configuration.fit.linearfitflag = 1
    configuration = pymca_configdict_to_model(
        mca_theory.configure(pymca_configdict_from_model(configuration))
    )
    fixed_spectrum_stdev = _set_reference_spectrum(
        mca_theory, xrf_spectra, configuration, individual_weights
    )

    # Apply channel limits
    channels, raw_channels = channel_info(mca_theory)
    num_fit_observations = len(channels)
    num_raw_observations = len(raw_channels)
    if num_fit_observations != num_raw_observations:
        xrf_spectra = numpy.take(xrf_spectra, channels, axis=1)
        if fixed_spectrum_stdev is not None:
            fixed_spectrum_stdev = numpy.take(fixed_spectrum_stdev, channels, axis=0)

    # Calculate the linear least-squares model matrix
    mca_theory.estimate()
    parameter_info_list = _free_linear_parameters(mca_theory)
    derivatives = _least_squares_model_matrix(
        mca_theory, parameter_info_list
    )  # shape (num_channels, num_lin_params)

    # Calculate the background of all XRF spectra
    anchor_indices = _background_anchor_indices(mca_theory, xrf_spectra, configuration)
    background = _calculate_background(xrf_spectra, configuration, anchor_indices)

    # Subtract background
    if background is not None:
        # Do not subtract in-place for the caller's sake and for type conversion
        xrf_spectra = xrf_spectra - background

    # Fit all XRF spectra
    peak_area_indices = [
        p.fit_parameter_index for p in parameter_info_list if p.is_peak_area
    ]
    positive_constraint = peak_area_indices if positive_peak_areas else None
    parameters, uncertainties, num_free_parameters = fit_profiles(
        derivatives,
        xrf_spectra.T,
        positive_constraint=positive_constraint,
        weight=configuration.fit.fitweight,
        profiles_stdev=fixed_spectrum_stdev,
    )

    # Model and residuals
    parameter_names = [p.name for p in parameter_info_list]

    if diagnostics:
        model = derivatives.dot(parameters).T  # shape (num_spectra, num_channels)
        residuals = xrf_spectra - model
        if background is not None:
            model += background
    else:
        spectra = None
        model = None
        residuals = None

    # Quantification
    if quantification:
        (
            mass_fractions,
            mass_fractions_per_layer,
            mass_fraction_names,
            quantification_parameters,
        ) = calculate_massfractions(
            mca_theory,
            parameters[peak_area_indices],
            [parameter_info_list[i].name for i in peak_area_indices],
            configuration,
            norm_factors,
        )
    else:
        mass_fractions = None
        mass_fractions_per_layer = None
        mass_fraction_names = None
        quantification_parameters = None

    # To be added in the future:
    # - error propagation of quantification
    # - areal density ng/mm² https://sourceforge.net/p/pymca/mailman/message/37039583/
    # - detection limits

    derivatives = derivatives.T  # shape (num_lin_params, num_channels)

    if len(channels) != num_raw_observations:
        # Expand with NaN's so the model, residuals and base profiles match to
        # orginal number of observations.
        if diagnostics:
            model = expand_data(channels, model, num_raw_observations, axis=1)
            residuals = expand_data(channels, residuals, num_raw_observations, axis=1)
        derivatives = expand_data(channels, derivatives, num_raw_observations, axis=1)

    energies = energy_info(mca_theory, channels)
    raw_energies = energy_info(mca_theory, raw_channels)

    return XRFBatchFitResult(
        parameters=parameters,
        uncertainties=uncertainties,
        derivatives=derivatives,
        parameter_names=parameter_names,
        channels=channels,
        energies=energies,
        raw_channels=raw_channels,
        raw_energies=raw_energies,
        spectra=spectra,
        model=model,
        residuals=residuals,
        mass_fractions=mass_fractions,
        mass_fractions_per_layer=mass_fractions_per_layer,
        mass_fraction_names=mass_fraction_names,
        quantification_parameters=quantification_parameters,
    )


class _FreeLinearParameter(NamedTuple):
    name: str
    parameter_index: int  # index in all parameters (linear/non-linear/fixed/free)
    fit_parameter_index: int  # index in free linear parameters
    is_peak_area: bool  # area of a peak group (e.g. Fe-K lines)


def _free_linear_parameters(mca_theory: ClassMcaTheory) -> List[_FreeLinearParameter]:
    parameters = []
    first_peak_area_index = mca_theory.NGLOBAL
    fit_parameter_index = 0
    for parameter_index, (name, code) in enumerate(
        zip(mca_theory.PARAMETERS, mca_theory.codes[0])
    ):
        if code == Gefit.CFIXED:
            continue
        # Not fixed means free and linear
        is_peak_area = parameter_index >= first_peak_area_index
        param = _FreeLinearParameter(
            name=name,
            parameter_index=parameter_index,
            fit_parameter_index=fit_parameter_index,
            is_peak_area=is_peak_area,
        )
        parameters.append(param)
        fit_parameter_index += 1
    return parameters


def _least_squares_model_matrix(
    mca_theory: ClassMcaTheory, parameter_info_list: List[_FreeLinearParameter]
) -> numpy.ndarray:
    """Returns shape (num_channels, num_free_param)."""
    derivatives = numpy.empty(
        (len(mca_theory.xdata), len(parameter_info_list)), dtype=numpy.float64
    )
    for param in parameter_info_list:
        param_deriv = mca_theory.linearMcaTheoryDerivative(
            mca_theory.parameters, param.parameter_index, mca_theory.xdata
        )
        derivatives[:, param.fit_parameter_index] = param_deriv
    return derivatives


def _calculate_background(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    anchor_indices: List[int],
) -> Optional[numpy.ndarray]:
    stripflag = configuration.fit.stripflag
    if not stripflag:
        return
    stripalgorithm = configuration.fit.stripalgorithm
    if stripalgorithm == 1:
        return _calculate_snip_background(xrf_spectra, configuration, anchor_indices)
    raise NotImplementedError(
        f"Background {stripalgorithm=} is not implemented for fast fitting"
    )


def _calculate_snip_background(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    anchor_indices: List[int],
) -> numpy.ndarray:

    if xrf_spectra.size == 0:
        return numpy.zeros_like(xrf_spectra)

    smooth_width = configuration.fit.stripfilterwidth
    snip_width = configuration.fit.snipwidth

    background = None
    for i, xrf_spectrum in enumerate(xrf_spectra):
        backgroundi = SpecfitFuns.SavitskyGolay(xrf_spectrum, smooth_width)
        last_anchor_index = 0
        for anchor_index in anchor_indices:
            if (anchor_index > last_anchor_index) and (anchor_index < backgroundi.size):
                backgroundi[last_anchor_index:anchor_index] = SpecfitFuns.snip1d(
                    backgroundi[last_anchor_index:anchor_index], snip_width, 0
                )
                last_anchor_index = anchor_index
        if last_anchor_index < backgroundi.size:
            backgroundi[last_anchor_index:] = SpecfitFuns.snip1d(
                backgroundi[last_anchor_index:], snip_width, 0
            )
        if i == 0:
            background = numpy.empty(dtype=backgroundi.dtype, shape=xrf_spectra.shape)
        background[i] = backgroundi
    return background


def _background_anchor_indices(
    mca_theory: ClassMcaTheory,
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
) -> List[int]:
    """Get anchors for background subtraction"""
    if not configuration.fit.stripflag:
        return []
    channels = mca_theory.xdata[:, 0]
    anchor_indices = []
    if configuration.fit.stripanchorsflag:
        if configuration.fit.stripanchorslist is not None:
            for channel in configuration.fit.stripanchorslist:
                if channel <= channels[0]:
                    continue
                index = numpy.nonzero(channels >= channel)[0]
                if len(index):
                    index = min(index)
                    if index > 0:
                        anchor_indices.append(index)
    if len(anchor_indices) == 0:
        anchor_indices = [0, xrf_spectra.shape[1] - 1]
    return sorted(anchor_indices)


def _set_reference_spectrum(
    mca_theory: ClassMcaTheory,
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    individual_weights: bool,
) -> Optional[numpy.ndarray]:
    """Returns the Least-Squares parameters for fit all XRF spectra."""
    fitweight = configuration.fit.fitweight
    fixed_spectrum_stdev = None

    if fitweight == 0:
        # No weights
        ref_spectrum = xrf_spectra[0]
        mca_theory.setData(y=ref_spectrum)
        return fixed_spectrum_stdev

    if fitweight == 1:
        # Poisson weights

        if individual_weights:
            # Per-pixel Poisson weights
            ref_spectrum = xrf_spectra[0]
            mca_theory.setData(y=ref_spectrum)
            return fixed_spectrum_stdev

        # Average Poisson weights
        sum_spectrum = xrf_spectra.sum(axis=0)
        mca_theory.setData(y=sum_spectrum)
        fixed_spectrum_stdev = numpy.sqrt(sum_spectrum) / len(xrf_spectra) + 1
        return fixed_spectrum_stdev

    raise NotImplementedError(
        f"Least-Squares parameter {fitweight=} is not implemented for fast fitting"
    )



def initializer():
    # Needs to happen before h5py is imported
    import os

    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

    import h5py

    return

    try:
        h5py._objects.phil.release()
    except Exception:
        pass

    h5py._objects.nonlocal_close()

    print("HDF5 OBJECTS pid = ", os.getpid())
    h5py._objects.print_reg()


def main(module_name: str, func_name: str, *arg, **kwargs):
    import importlib

    mod = importlib.import_module(module_name)
    func = getattr(mod, func_name)

    return func(*arg, **kwargs)



from typing import List

from AnyQt import QtWidgets
from ewoksorange.gui.owwidgets.meta import ow_build_opts
from ewoksorange.gui.owwidgets.nothread import OWEwoksWidgetNoThread

from ewoksfluo.gui.data_viewer import DataViewer
from ewoksfluo.gui.pick_scans_widget import PickScansWidget
from ewoksfluo.tasks.input.pick_scans import PickScans


class OWPickScans(OWEwoksWidgetNoThread, **ow_build_opts, ewokstaskclass=PickScans):
    name = "Pick scans"
    description = "Pick scans"

    def __init__(self):
        super().__init__()
        self.tabs: List[PickScansWidget] = []
        self._init_control_area()
        self._init_main_area()

    @property
    def n_tabs(self) -> int:
        return len(self.tabs)

    @property
    def plus_index(self) -> int:
        return self.n_tabs

    @property
    def minus_index(self) -> int:
        return self.n_tabs + 1

    def handle_special_tabs(self, i) -> None:
        """
        Handles clicks on special tabs (+ or -)
        """
        if i == self.plus_index:
            self.add_new_picking_tab()
            return

        if i == self.minus_index:
            self.remove_last_picking_tab()
            return

    def add_new_picking_tab(self, initial_values: dict = {}) -> None:
        # Insert the new tab before the + index
        new_tab_index = self.plus_index
        page_widget = PickScansWidget(
            lambda inputs: self.update_tab_inputs(inputs, new_tab_index),
            initial_values,
        )
        self.tabWidget.insertTab(new_tab_index, page_widget, f"File {new_tab_index+1}")
        self.tabs.insert(new_tab_index, page_widget)

        self.tabWidget.setCurrentIndex(new_tab_index)

    def remove_last_picking_tab(self) -> None:
        last_tab_index = self.plus_index - 1
        self.tabWidget.setCurrentIndex(last_tab_index)
        if last_tab_index == 0:
            return
        self.tabWidget.removeTab(last_tab_index)
        w = self.tabs.pop()
        filename = w.form.get_parameter_value("filename")
        if filename:
            self._viewer.closeFile(filename)
        del w
        self.update_inputs()

    def require_number_of_tabs(self, n: int):
        while self.n_tabs < max(n, 1):
            self.add_new_picking_tab()
        while self.n_tabs > max(n, 1):
            self.remove_last_picking_tab()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._viewer = DataViewer(parent=self.mainArea)
        self._viewer.setVisible(True)
        layout.addWidget(self._viewer)
        layout.setStretchFactor(self._viewer, 1)

        self._refresh_files()

    def _init_control_area(self) -> None:
        super()._init_control_area()
        self.tabWidget = QtWidgets.QTabWidget(parent=self.controlArea)
        self.tabWidget.addTab(QtWidgets.QWidget(), "+")
        self.tabWidget.addTab(QtWidgets.QWidget(), "-")
        self.tabWidget.tabBar().setSelectionBehaviorOnRemove(
            QtWidgets.QTabBar.SelectionBehavior.SelectLeftTab
        )
        self.tabWidget.currentChanged.connect(lambda i: self.handle_special_tabs(i))

        self._get_control_layout().addWidget(self.tabWidget)

        initial_values = self.get_default_input_values(include_missing=False)

        if not initial_values:
            self.add_new_picking_tab()
        else:
            for filename, scan_range, exclude_scans in zip(
                initial_values["filenames"],
                initial_values["scan_ranges"],
                initial_values["exclude_scans"],
            ):
                scan_min, scan_max = scan_range if scan_range else (None, None)
                self.add_new_picking_tab(
                    {
                        "filename": filename,
                        "scan_min": scan_min,
                        "scan_max": scan_max,
                        "exclude_scans": exclude_scans,
                    }
                )

    def handleNewSignals(self) -> None:
        self._update_input_data()
        self._refresh_files()
        super().handleNewSignals()

    def _update_input_data(self) -> None:
        values = self.get_task_input_values()
        self.require_number_of_tabs(len(values.get("filenames", list())))

        dynamic = self.get_dynamic_input_names()
        parameters = {
            "filenames": ["filename"],
            "scan_ranges": ["scan_min", "scan_max"],
            "exclude_scans": ["exclude_scans"],
        }
        for name in self.get_input_names():
            for w in self.tabs:
                for pname in parameters[name]:
                    w.form.set_parameter_enabled(pname, name not in dynamic)

    def update_inputs(self):
        for i, tab in enumerate(self.tabs):
            self.update_tab_inputs(tab.form.get_parameter_values(), i)

    def update_tab_inputs(self, form_values: dict, index: int) -> None:
        values = self.get_default_input_values(include_missing=True)

        # Make values index match the tabs index
        missing_items = {"filenames": "", "scan_ranges": [1, 1], "exclude_scans": []}
        for name in self.get_input_names():
            if not values.get(name, None):
                values[name] = []
            if len(values[name]) < len(self.tabs):
                values[name].extend(
                    [missing_items[name]] * (len(self.tabs) - len(values[name]))
                )
            if len(values[name]) > len(self.tabs):
                values[name] = values[name][: len(self.tabs)]

        # Update only if the new value is valid (i.e. not missing)
        if form_values["filename"]:
            values["filenames"][index] = form_values["filename"]
        if form_values["scan_min"] and form_values["scan_max"]:
            values["scan_ranges"][index] = [
                form_values["scan_min"],
                form_values["scan_max"],
            ]
        if form_values["exclude_scans"]:
            values["exclude_scans"][index] = form_values["exclude_scans"]

        self.update_default_inputs(**values)
        self._refresh_files()

    def _refresh_files(self):
        filenames = self.get_task_input_value("filenames", [])
        for filename in filenames:
            if not filename:
                continue
            self._viewer.updateFile(filename)

    def _execute_ewoks_task(self, *args, **kw) -> None:
        self._viewer.closeAll()
        super()._execute_ewoks_task(*args, **kw)

    def task_output_changed(self):
        self._refresh_files()



import numpy

from ..tasks.example_data.deadtime import apply_dualchannel_signal_processing


def test_apply_dualchannel_signal_processing():
    for _ in range(5):
        measured = apply_dualchannel_signal_processing(
            numpy.array([[10000]]), elapsed_time=1.0
        )
        counts = measured["spectrum"] / measured["live_time"] * 0.1
        if counts[0, 0] != 10000:
            break
    else:
        assert counts[0, 0] != 10000

    measured = apply_dualchannel_signal_processing(
        numpy.array([[10000.1]]),
        elapsed_time=0.1,
        counting_noise=False,
        integral_type=numpy.uint32,
    )
    counts = measured["spectrum"] / measured["live_time"] * 0.1
    assert counts[0, 0] == 10000

    measured = apply_dualchannel_signal_processing(
        numpy.array([[10000.1]]),
        elapsed_time=0.1,
        counting_noise=False,
        integral_type=None,
    )
    counts = measured["spectrum"] / measured["live_time"] * 0.1
    numpy.testing.assert_almost_equal(counts[0, 0], 10000.1, decimal=9)



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy

from ..model import PyMcaXrfConfiguration
from .matrix import ensure_matrix

logger = logging.getLogger(__name__)


def set_beam_energy(
    configuration: PyMcaXrfConfiguration,
    energy: Optional[float],
    energy_multiplier: Optional[float],
):
    parameters = _get_beam_energy_parameters(configuration, energy, energy_multiplier)
    if parameters:
        _apply_beam_energy_parameters(configuration, parameters)


def _get_beam_energy_parameters(
    configuration: PyMcaXrfConfiguration,
    energy: Optional[float],
    energy_multiplier: Optional[float],
) -> Dict[str, List[Union[float, int]]]:
    if not _is_positive_finite_number(energy) and energy_multiplier is None:
        return {}

    fit = configuration.fit
    old_cfg = {
        "energy": _as_list(fit.energy),
        "energyweight": _as_list(fit.energyweight),
        "energyflag": _as_list(fit.energyflag),
        "energyscatter": _as_list(fit.energyscatter),
    }

    valid_items = [
        (i, e)
        for i, (e, energyflag) in enumerate(
            zip(old_cfg["energy"], old_cfg["energyflag"])
        )
        if energyflag == 1 and _is_positive_finite_number(e)
    ]
    indices = [i for i, _ in sorted(valid_items, key=lambda tpl: tpl[1])]

    modified = False

    if indices:

        def extract_defined_energies(name: str, dtype):
            energies = [old_cfg[name][i] for i in indices]
            return numpy.array(energies, dtype=dtype).tolist()

        parameters: Dict[str, list] = {
            "energy": extract_defined_energies("energy", float),
            "energyflag": extract_defined_energies("energyflag", int),
            "energyweight": extract_defined_energies("energyweight", float),
            "energyscatter": extract_defined_energies("energyscatter", int),
        }

        if _is_positive_finite_number(energy):
            first_energy = parameters["energy"][0]
            if not numpy.isclose(first_energy, energy):
                # Change the first energy and renormalize the others
                parameters["energy"] = [
                    e * energy / first_energy for e in parameters["energy"]
                ]
                modified = True

                logger.warning(
                    "PyMca configuration: modify primary beam energies to %s keV",
                    parameters["energy"],
                )
    else:

        if not _is_positive_finite_number(energy):
            return {}

        parameters: Dict[str, list] = {
            "energy": [energy],
            "energyflag": [1],
            "energyweight": [1.0],
            "energyscatter": [0],
        }
        modified = True

        logger.warning("PyMca configuration: set primary beam energy to %s keV", energy)

    # Add faraway line at the end (with small weight to not be fitted)
    # to be sure to include all lines in the energy range.
    first_energy = parameters["energy"][0]
    last_energy = parameters["energy"][-1]
    if _is_positive_finite_number(energy_multiplier):
        extra_energy = first_energy * energy_multiplier
        if last_energy < extra_energy:
            parameters["energy"].append(extra_energy)
            parameters["energyflag"].append(1)
            parameters["energyweight"].append(1e-10)
            parameters["energyscatter"].append(0)
            modified = True

            logger.warning(
                "PyMca configuration: append primary beam energy at %s keV (negligible contribution)",
                extra_energy,
            )

    if modified:
        return parameters
    return {}


def _apply_beam_energy_parameters(
    configuration: PyMcaXrfConfiguration,
    parameters: Dict[str, List[Union[float, int]]],
):
    fit = configuration.fit
    for key, value in parameters.items():
        setattr(fit, key, value)

    # Dummy matrix (apparently needed for multi-energy)
    if len(parameters["energy"]) > 1:
        ensure_matrix(configuration)


def _as_list(value: Any) -> List[Any]:
    if isinstance(value, list):
        return value
    return [value]


def _is_positive_finite_number(value: Any) -> bool:
    if not isinstance(value, (int, float)):
        return False
    return value > 0 and numpy.isfinite(value)



import re
from typing import Any
from typing import Dict

import h5py
import numpy

from ..io import hdf5
from . import nexus_utils

_PEAKS_AREA_REGEX = [
    re.compile(r"^[a-zA-Z]+_[KLM][a-b1-5]?$"),
    re.compile(r"^Scatter_Compton[0-9]+$"),
    re.compile(r"^Scatter_Peak[0-9]+$"),
]


def is_peak_area(dset: hdf5.DatasetType) -> bool:
    """Checks if the dataset is a peak area."""
    if not hdf5.is_dataset(dset):
        return False
    dset_name = dset.name.split("/")[-1]
    return any(regex.match(dset_name) for regex in _PEAKS_AREA_REGEX)


def save_xrf_results(
    output_root_uri: str,
    group_name: str,
    process_config: Dict[str, Any],
    parameters: Dict[str, numpy.ndarray],
    uncertainties: Dict[str, numpy.ndarray],
    massfractions: Dict[str, numpy.ndarray],
) -> str:
    start_time = nexus_utils.now()
    with nexus_utils.save_in_ewoks_process(
        output_root_uri,
        start_time,
        process_config=process_config,
        default_levels=("results", group_name),
    ) as (process_group, already_existed):
        if already_existed:
            results_group = process_group["results"]
        else:
            results_group = process_group.create_group("results")
            results_group.attrs["NX_class"] = "NXcollection"

            if parameters:
                _ = _save_nxdata(results_group, "parameters", parameters)
            if uncertainties:
                _ = _save_nxdata(results_group, "uncertainties", uncertainties)
            # if parameters and uncertainties:
            #    for name in set(param_group) & set(error_group):
            #        create_hdf5_link(param_group, f"{name}_errors", error_group[name])
            if massfractions:
                _save_nxdata(results_group, "massfractions", massfractions)
        return f"{results_group.file.filename}::{results_group.name}"


def _save_nxdata(parent: hdf5.GroupType, name: str, group: Dict[str, numpy.ndarray]):
    nxgroup = nexus_utils.create_nxdata(parent, name)
    for name, data in group.items():
        nxgroup.create_dataset(name, data=data)
    nexus_utils.set_nxdata_signals(nxgroup, signals=list(group))
    return nxgroup


def get_xrf_result_groups(parent_nxdata: hdf5.GroupType) -> Dict[str, h5py.Group]:
    """Most important group comes first"""
    return {
        name: parent_nxdata[name] for name in _NXDATA_ORDER if name in parent_nxdata
    }


_NXDATA_ORDER = [
    "parameters",
    "massfractions",
    "uncertainties",
    "rawcounters",
]



from typing import Optional

import numpy
from PyMca5.PyMcaPhysics.xrf.FastXRFLinearFit import FastXRFLinearFit
from PyMca5.PyMcaPhysics.xrf.LegacyFastXRFLinearFit import (
    FastXRFLinearFit as LegacyFastXRFLinearFit,
)

from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import pymca_configdict_from_model
from ._utils.parse_native import parse_native_fit_output
from ._utils.parse_native import parse_native_legacy_fit_output
from .types import XRFBatchFitResult


def fit_xrf_spectra(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Fast fitting refers to linear fitting of several spectra by solving a single
    system of equations. This function uses the high-level API from PyMca.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    fastFit = FastXRFLinearFit()

    if configuration.fit.fitweight:
        weight = 1 + bool(individual_weights)
    else:
        weight = 0

    # WARNING: FastXRFLinearFit ignores `use_limit=False`
    #          and always uses xmin/xmax from the configuration.
    if configuration.fit.use_limit:
        xmin = None
        xmax = None
    else:
        xmin = 0
        xmax = xrf_spectra.shape[1]

    with numpy.errstate(over="ignore"):  # Stores results in float32
        outbuffer = fastFit.fitMultipleSpectra(
            y=xrf_spectra.copy(),  # background subtraction in-place
            refit=int(positive_peak_areas),
            concentrations=int(quantification),
            livetime=live_times,
            weight=weight,
            configuration=pymca_configdict_from_model(configuration),
            saveResiduals=True,
            saveFit=True,
            saveData=True,
            diagnostics=True,
            saveFOM=True,
            xmin=xmin,
            xmax=xmax,
        )

    return parse_native_fit_output(
        xrf_spectra,
        fastFit._mcaTheory,
        outbuffer,
        quantification,
        diagnostics,
        fast_fitting=True,
    )


def fit_xrf_spectra_legacy(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
) -> XRFBatchFitResult:
    """Fast fitting refers to linear fitting of several spectra by solving a single
    system of equations. This function uses the legacy high-level API from PyMca.

    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    """

    if quantification is None:
        quantification = True
    if individual_weights is None:
        individual_weights = True
    if positive_peak_areas is None:
        positive_peak_areas = True
    if diagnostics is None:
        diagnostics = False

    fastFit = LegacyFastXRFLinearFit()

    if configuration.fit.fitweight:
        weight = 1 + bool(individual_weights)
    else:
        weight = 0

    # WARNING: FastXRFLinearFit ignores `use_limit=False`
    #          and always uses xmin/xmax from the configuration.
    if configuration.fit.use_limit:
        xmin = None
        xmax = None
    else:
        xmin = 0
        xmax = xrf_spectra.shape[1]

    y = xrf_spectra.copy()[numpy.newaxis, ...]  # background subtraction in-place

    with numpy.errstate(over="ignore"):  # Stores results in float32
        outputdict = fastFit.fitMultipleSpectra(
            y=y,
            refit=int(positive_peak_areas),
            concentrations=int(quantification),
            livetime=live_times,
            weight=weight,
            configuration=pymca_configdict_from_model(configuration),
            xmin=xmin,
            xmax=xmax,
        )

    return parse_native_legacy_fit_output(
        xrf_spectra,
        fastFit._mcaTheory,
        outputdict,
        quantification,
        diagnostics,
        fast_fitting=True,
    )



from typing import List
from typing import Optional

import numpy

from ..model import PyMcaXrfConfiguration


def configuration_summary(
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[dict],
    fast_fitting: bool,
) -> List[str]:
    fit = configuration.fit
    concentrations = configuration.concentrations
    attenuators = configuration.attenuator_views
    matrix = attenuators["Matrix"]

    ind = numpy.array(fit.energyflag).astype(bool)
    _energy = numpy.array(fit.energy)[ind]
    _weights = numpy.array(fit.energyweight)[ind]
    _weights = _weights / _weights.sum() * 100
    _scatter = numpy.array(fit.energyscatter)[ind]

    summary: List[str] = [
        f"{en} keV (Rate = {w:.2f}%, Scatter {'ON' if scat else 'OFF'})"
        for en, w, scat in zip(_energy, _weights, _scatter)
    ]

    if quantification:
        summary.extend(
            [
                f"flux = {concentrations.flux:e} ph/sec",
                f"time = {concentrations.time} s",
                f"active area = {concentrations.area} cm^2",
                f"sample-detector distance = {concentrations.distance} cm",
                f"angle IN = {matrix.incidence_angle} deg",
                f"angle OUT = {matrix.exit_angle} deg",
            ]
        )

    if matrix.enabled:
        summary.append(f"Matrix = {matrix.material}")
    else:
        summary.append("Matrix = <DISABLED>")

    summary.append(f"Linear = {'YES' if fit.linearfitflag else 'NO'}")
    summary.append(f"Fast fitting = {'YES' if fast_fitting else 'NO'}")
    summary.append(f"Error propagation = {'Poisson' if fit.fitweight else 'OFF'}")
    summary.append(f"Matrix adjustment = {'ON' if fit.strategyflag else 'OFF'}")

    return summary



import re

import numpy
from PyMca5.PyMcaPhysics.xrf import ClassMcaTheory
from PyMca5.PyMcaPhysics.xrf.XRFBatchFitOutput import OutputBuffer

from ..types import NotAvailable
from ..types import XRFBatchFitResult
from .calibration import channel_info
from .calibration import energy_info


def parse_native_fit_output(
    xrf_spectra: numpy.ndarray,
    mca_theory: ClassMcaTheory,
    outbuffer: OutputBuffer,
    quantification: bool,
    diagnostics: bool,
    fast_fitting: bool,
) -> XRFBatchFitResult:
    """Convert native PyMca batch fitting result to `XRFBatchFitResult`."""
    channels, raw_channels = channel_info(mca_theory)
    if not fast_fitting:
        start_channel = numpy.round(channels[0]).astype(int)
        stop_channel = numpy.round(channels[-1]).astype(int) + 1
        channels = raw_channels[start_channel:stop_channel]

    energies = energy_info(mca_theory, channels)
    raw_energies = energy_info(mca_theory, raw_channels)

    parameters = outbuffer["parameters"]
    uncertainties = outbuffer["uncertainties"]
    parameter_names = outbuffer.labels("parameters")

    if quantification:
        massfractions_merged = outbuffer["massfractions"]

        mass_fraction_names = []
        layer_indices = {}
        for idx, name in enumerate(outbuffer.labels("massfractions")):
            if isinstance(name, str):
                mass_fraction_names.append(name)
                key = 0
            else:
                _, layer = name
                key = int(layer.lower().split("layer")[1])
            layer_indices.setdefault(key, []).append(idx)

        if len(layer_indices) == 1:
            mass_fractions = massfractions_merged
            mass_fractions_per_layer = []
        else:
            mass_fractions = massfractions_merged[layer_indices[0]]
            mass_fractions_per_layer = [
                massfractions_merged[layer_indices[key]]
                for key in sorted(layer_indices)[1:]
            ]
    else:
        mass_fractions = None
        mass_fraction_names = None
        mass_fractions_per_layer = None

    quantification_parameters = NotAvailable()
    if diagnostics:
        spectra = xrf_spectra
        model = outbuffer["model"]
        residuals = outbuffer["residuals"]
    else:
        spectra = None
        model = None
        residuals = None

    if fast_fitting:
        derivatives = outbuffer["derivatives"]
    else:
        derivatives = None

        parameters = _flatten_3d_last(parameters)
        uncertainties = _flatten_3d_last(uncertainties)
        if quantification:
            mass_fractions = _flatten_3d_last(mass_fractions)
            mass_fractions_per_layer = [
                _flatten_3d_last(arr) for arr in mass_fractions_per_layer
            ]
        if diagnostics:
            model = _flatten_3d_first(model)
            residuals = _flatten_3d_first(residuals)

    return XRFBatchFitResult(
        parameters=parameters,
        uncertainties=uncertainties,
        derivatives=derivatives,
        parameter_names=parameter_names,
        channels=channels,
        energies=energies,
        raw_channels=raw_channels,
        raw_energies=raw_energies,
        spectra=spectra,
        model=model,
        residuals=residuals,
        mass_fractions=mass_fractions,
        mass_fractions_per_layer=mass_fractions_per_layer,
        mass_fraction_names=mass_fraction_names,
        quantification_parameters=quantification_parameters,
    )


def parse_native_legacy_fit_output(
    xrf_spectra: numpy.ndarray,
    mca_theory: ClassMcaTheory,
    outputdict: dict,
    quantification: bool,
    diagnostics: bool,
    fast_fitting: bool,
) -> XRFBatchFitResult:
    """Convert native PyMca legacy batch fitting result to `XRFBatchFitResult`."""
    channels, raw_channels = channel_info(mca_theory)
    energies = energy_info(mca_theory, channels)
    raw_energies = energy_info(mca_theory, raw_channels)

    parameters = _flatten_3d_last(outputdict["parameters"])
    uncertainties = _flatten_3d_last(outputdict["uncertainties"])
    names = outputdict["names"]
    if quantification:
        noffset = len(outputdict["parameters"])
        parameter_names = names[:noffset]

        massfractions_merged = _flatten_3d_last(outputdict["concentrations"])

        mass_fraction_names = []
        layer_indices = {}
        for idx, name in enumerate(names[noffset:]):
            if "layer" not in name.lower():
                key = 0
                name = re.search(r"\((.*?)\)", name).group(1)
                mass_fraction_names.append(name)
            else:
                key = int(name.lower().split("-layer")[1])
            layer_indices.setdefault(key, []).append(idx)

        if len(layer_indices) == 1:
            mass_fractions = massfractions_merged
            mass_fractions_per_layer = []
        else:
            mass_fractions = massfractions_merged[layer_indices[0]]
            mass_fractions_per_layer = [
                massfractions_merged[layer_indices[key]]
                for key in sorted(layer_indices)[1:]
            ]

    else:
        parameter_names = names
        mass_fractions = None
        mass_fraction_names = None
        mass_fractions_per_layer = None

    if diagnostics:
        spectra = xrf_spectra
    else:
        spectra = None

    return XRFBatchFitResult(
        parameters=parameters,
        uncertainties=uncertainties,
        derivatives=NotAvailable(),
        parameter_names=parameter_names,
        channels=channels,
        energies=energies,
        raw_channels=raw_channels,
        raw_energies=raw_energies,
        spectra=spectra,
        model=NotAvailable(),
        residuals=NotAvailable(),
        mass_fractions=mass_fractions,
        mass_fractions_per_layer=mass_fractions_per_layer,
        mass_fraction_names=mass_fraction_names,
        quantification_parameters=NotAvailable(),
    )


def _flatten_3d_last(arr: numpy.ndarray) -> numpy.ndarray:
    """Reshape array with shape (A, B, C) into shape (A, B*C)"""
    if arr.ndim != 3:
        raise ValueError(f"Numpy is {arr.ndim}D instead of 3D")
    return arr.reshape(arr.shape[0], -1)


def _flatten_3d_first(arr: numpy.ndarray) -> numpy.ndarray:
    """Reshape array with shape (A, B, C) into shape (A*B, C)"""
    if arr.ndim != 3:
        raise ValueError(f"Numpy is {arr.ndim}D instead of 3D")
    return arr.reshape(-1, arr.shape[-1])



import functools
import os
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple

import h5py
import numpy
from silx.io import h5py_utils

try:
    from imageio.v3 import imread
except ImportError:
    try:
        from imageio.v2 import imread
    except ImportError:
        from imageio import imread

from scipy import ndimage

from ...io import hdf5
from ...xrffit.pymca_config import PyMcaXrfConfiguration
from ...xrffit.pymca_config import pymca_configdict_from_model
from .. import nexus_utils
from . import xrf_spectra
from .deadtime import apply_dualchannel_signal_processing
from .monitor import monitor_signal


def save_2d_xrf_scans(
    filename: str,
    emission_line_groups: List[str],
    first_scan_number: int,
    shape: Tuple[int, int],
    mosaic: Tuple[int, int],
    energy: float = 12,
    flux: float = 1e7,
    expo_time: float = 0.1,
    counting_noise: bool = True,
    integral_type: bool = True,
    rois: Sequence = tuple(),
    nmcas: int = 1,
    max_deviation: float = 0,
    seed: Optional[int] = None,
) -> List[int]:
    """Simulates and saves 2D XRF scans to a NeXus file.

    :param filename: Output file path.
    :param emission_line_groups: List of emission line groups as "Element-Group".
    :param first_scan_number: Scan number of the first scan in the output file.
    :param shape: Scan shape.
    :param mosaic: Split shape in this number of blocks.
    :param energy: Incident X-ray energy in keV.
    :param flux: Incident X-ray flux in photons per second.
    :param expo_time: Exposure time per point in seconds.
    :param counting_noise: If True, adds Poisson noise to data.
    :param integral_type: If True, data represents integrated intensities.
    :param rois: Regions of interest as energy ranges.
    :param ndetectors: Number of detectors.
    :param max_deviation: Maximum deviation from a perfect grid as a fraction of the scan step size.
    :param seed: Random seed for reproducibility.

    :returns: Scan numbers as many as there are blocks in the mosaic.

    :raises ValueError: If `shape` is not 2D.
    """
    max_deviation = max(max_deviation, 0)
    rstate = numpy.random.RandomState(seed=seed)
    I0_max = int(flux * expo_time)
    emission_line_groups = [s.split("-") for s in emission_line_groups]

    specs = _amesh_specs(
        shape=shape, mosaic=mosaic, max_deviation=max_deviation, expo_time=expo_time
    )
    scan_numbers = list()
    for scan_number, amesh in enumerate(specs, start=first_scan_number):
        scan_numbers.append(scan_number)

        coordinates = _amesh_positions(amesh, rstate, max_deviation)

        # decaying I0 image
        I0 = (I0_max * monitor_signal(expo_time, amesh.size)).reshape(
            amesh.shape, order="F"
        )

        # random images with values between 0 and 1
        fluoI0fractions = list(
            _iter_amesh_data(
                amesh, coordinates, rstate, nmaps=len(emission_line_groups)
            )
        )
        scatterI0fractions = list(_iter_amesh_data(amesh, coordinates, rstate, nmaps=2))

        # Peak area counts within expo_time seconds
        linegroups = [
            xrf_spectra.EmissionLineGroup(
                element, group, (I0 * I0fraction).astype(numpy.uint32)
            )
            for I0fraction, (element, group) in zip(
                fluoI0fractions, emission_line_groups
            )
        ]
        scattergroups = [
            xrf_spectra.ScatterLineGroup(
                "Compton000", (I0 * scatterI0fractions[0]).astype(numpy.uint32)
            ),
            xrf_spectra.ScatterLineGroup(
                "Peak000", (I0 * scatterI0fractions[1]).astype(numpy.uint32)
            ),
        ]

        # Theoretical XRF spectra
        theoretical_spectra, config = xrf_spectra.xrf_spectra(
            linegroups,
            scattergroups,
            energy=energy,
            flux=flux,
            elapsed_time=expo_time,
        )

        # Measured XRF spectra
        if integral_type:
            integral_type = numpy.uint32
        else:
            integral_type = None
        measured_data = apply_dualchannel_signal_processing(
            theoretical_spectra,
            elapsed_time=expo_time,
            counting_noise=counting_noise,
            integral_type=integral_type,
        )

        # ROI data (theoretical, measured and corrected)
        roi_data_theory = dict()
        roi_data_cor = dict()  # I0 and LT corrected
        I0_reference = I0_max
        for i, roi in enumerate(rois, 1):
            roi_name = f"roi{i}"
            idx = Ellipsis, slice(*roi)

            roi_theory = theoretical_spectra[idx].sum(axis=-1) / I0 * I0_max
            roi_data_theory[roi_name] = roi_theory

            roi_meas = measured_data["spectrum"][idx].sum(axis=-1)
            measured_data[roi_name] = roi_meas

            cor = I0_reference / I0 * expo_time / measured_data["live_time"]
            roi_data_cor[roi_name] = cor * roi_meas

        _save_2d_xrf_scans(
            filename,
            scan_number,
            amesh,
            coordinates,
            energy,
            I0,
            I0_reference,
            nmcas,
            measured_data,
            roi_data_theory,
            roi_data_cor,
            linegroups,
            scattergroups,
            config,
        )
    return scan_numbers


class _AmeshAxis(NamedTuple):
    name: str
    start: float
    end: float
    intervals: int

    @property
    def size(self) -> int:
        return self.intervals + 1

    @property
    def coordinates(self) -> numpy.ndarray:
        return numpy.linspace(self.start, self.end, self.size)


class _Amesh(NamedTuple):
    fast: _AmeshAxis
    slow: _AmeshAxis
    expo_time: float

    def __str__(self):
        return f"amesh {self.fast.name} {self.fast.start} {self.fast.end} {self.fast.intervals} {self.slow.name} {self.slow.start} {self.slow.end} {self.slow.intervals} {self.expo_time}"

    @property
    def shape(self) -> Tuple[int]:
        return (self.fast.size, self.slow.size)

    @property
    def size(self) -> int:
        return self.fast.size * self.slow.size


@functools.lru_cache(maxsize=1)
def _scene_data() -> Tuple[numpy.ndarray, List[str]]:
    """Data in pixel coordinates on which to interpolate for create amesh data.

    :returns: array with shape `(nimages, nfast, nslow)`, fast and slow axis name (in that order)
    """
    filename = os.path.join(os.path.dirname(__file__), "ihc.png")
    scene_data = numpy.transpose(imread(filename), [2, 0, 1])
    _mark_fast_axis(scene_data)
    scene_axes = ["sampz", "sampy"]
    return scene_data, scene_axes


def _mark_fast_axis(channels: numpy.ndarray) -> None:
    """Modify the image intensities to mark the fast axis (first dimension)."""
    image_shape = channels[0].shape

    dstart = image_shape[0] // 10, image_shape[1] // 10
    dtick = dstart[0] // 4, dstart[1] // 4

    p0 = dstart[0] - dtick[0], dstart[1] - dtick[1]
    p1 = dstart[0] + dtick[0], dstart[1] + dtick[1]
    channels[:, p0[0] : p1[0], p0[1] : p1[1]] = 255

    dtick = dtick[0] // 2, dtick[1] // 2
    dend = image_shape[0] // 2, image_shape[1] // 10

    p0 = dstart[0] - dtick[0], dstart[1] - dtick[1]
    p1 = dend[0] + dtick[0], dend[1] + dtick[1]
    channels[:, p0[0] : p1[0], p0[1] : p1[1]] = 255


def _amesh_specs(
    shape: Tuple[int, int],
    mosaic: Tuple[int, int],
    max_deviation: float = 0,
    expo_time: float = 0.1,
) -> List[_Amesh]:
    """
    Generate a list of _Amesh instances based on the provided parameters.

    :param shape: Number of interpolation points along each axis (fast, slow).
    :param mosaic: Number of blocks along each axis (fast, slow).
    :param max_deviation: Maximum deviation from a perfect grid as a fraction of the scan step size.
    :param expo_time: Exposure time per point in seconds.
    :return: List of _Amesh instances.
    """
    scene_data, scene_axes = _scene_data()
    scene_shape = scene_data.shape[1:]

    fast_axes = _amesh_axes(
        scene_axes[0], scene_shape[0], shape[0], mosaic[0], max_deviation
    )
    slow_axes = _amesh_axes(
        scene_axes[1], scene_shape[1], shape[1], mosaic[1], max_deviation
    )
    amesh_list = list()
    for slow in slow_axes:
        for fast in fast_axes:
            amesh_list.append(_Amesh(fast, slow, expo_time))
    return amesh_list


def _amesh_axes(
    name: str, scene_size: int, total_size: int, mosaic: int, max_deviation: float
) -> List[_AmeshAxis]:
    """
    Generate a list of _AmeshAxis instances based on the provided parameters.

    :param name: Axis name.
    :param scene_size: Total of scene data points.
    :param total_size: Number of scan points.
    :param mosaic: Number of scans to span the scene.
    :param max_deviation: Maximum deviation from a perfect grid as a fraction of the scan step size.
    """
    block_size = total_size // mosaic
    block_intervals = block_size - 1

    scene_block_range = (scene_size - 1) / (
        mosaic + (2 * max_deviation / block_intervals)
    )
    interpolate_step_size = scene_block_range / block_intervals
    scan_border = max_deviation * interpolate_step_size
    scene_start = scan_border
    # scene_end = scene_size - 1 - scan_border
    # assert ((scene_end - scene_start) / scene_block_range - mosaic) < 1e-6

    axes = list()
    for i in range(mosaic):
        block_start = scene_start + i * scene_block_range
        block_end = block_start + scene_block_range

        axes.append(
            _AmeshAxis(
                name=name,
                start=block_start,
                end=block_end,
                intervals=block_intervals,
            )
        )
    return axes


def _save_2d_xrf_scans(
    filename: str,
    scan_number: int,
    amesh: _Amesh,
    coordinates: List[numpy.ndarray],
    energy: float,
    I0: numpy.ndarray,
    I0_reference: float,
    nmcas: int,
    measured_data: Dict[str, numpy.ndarray],
    roi_data_theory: Dict[str, numpy.ndarray],
    roi_data_cor: Dict[str, numpy.ndarray],
    linegroups: List[xrf_spectra.EmissionLineGroup],
    scattergroups: List[xrf_spectra.ScatterLineGroup],
    pymca_configuration: PyMcaXrfConfiguration,
) -> None:
    with h5py_utils.File(filename, mode="a") as nxroot:
        scan_name = f"{scan_number}.1"
        nxroot.attrs["NX_class"] = "NXroot"
        nxroot.attrs["creator"] = "ewoksfluo"

        nxentry = nxroot.require_group(scan_name)
        nxentry.attrs["NX_class"] = "NXentry"
        if "title" in nxentry:
            del nxentry["title"]
        nxentry["title"] = str(amesh)

        nxinstrument = nxentry.require_group("instrument")
        nxinstrument.attrs["NX_class"] = "NXinstrument"

        measurement = nxentry.require_group("measurement")
        measurement.attrs["NX_class"] = "NXcollection"

        fast_coordinates = coordinates[0].flatten(order="F")
        slow_coordinates = coordinates[1].flatten(order="F")

        # Positioners
        for name in ("positioners_start", "positioners_end", "positioners"):
            group = nxinstrument.require_group(name)
            group.attrs["NX_class"] = "NXcollection"

            if "energy" in group:
                del group["energy"]
            group["energy"] = energy
            group["energy"].attrs["units"] = "keV"

            if name == "positioners":
                continue

            if name == "positioners_start":
                idx = 0
            else:
                idx = -1

            if amesh.fast.name in group:
                del group[amesh.fast.name]
            group[amesh.fast.name] = fast_coordinates[idx]
            group[amesh.fast.name].attrs["units"] = "um"

            if amesh.slow.name in group:
                del group[amesh.slow.name]
            group[amesh.slow.name] = slow_coordinates[idx]
            group[amesh.slow.name].attrs["units"] = "um"
        positioners = nxinstrument["positioners"]

        # I0 data
        nxdetector = nxinstrument.require_group("I0")
        nxdetector.attrs["NX_class"] = "NXdetector"
        if "data" in nxdetector:
            del nxdetector["data"]
        nxdetector["data"] = I0.flatten(order="F")
        if "I0" not in measurement:
            measurement["I0"] = h5py.SoftLink(nxdetector["data"].name)

        # Fast axis
        nxpositioner = nxinstrument.require_group(amesh.fast.name)
        nxpositioner.attrs["NX_class"] = "NXpositioner"
        if "value" in nxpositioner:
            del nxpositioner["value"]
        nxpositioner["value"] = fast_coordinates
        nxpositioner["value"].attrs["units"] = "um"
        if amesh.fast.name not in measurement:
            measurement[amesh.fast.name] = h5py.SoftLink(nxpositioner["value"].name)
        if amesh.fast.name not in positioners:
            positioners[amesh.fast.name] = h5py.SoftLink(nxpositioner["value"].name)

        # Slow axis
        nxpositioner = nxinstrument.require_group(amesh.slow.name)
        nxpositioner.attrs["NX_class"] = "NXpositioner"
        if "value" in nxpositioner:
            del nxpositioner["value"]
        nxpositioner["value"] = slow_coordinates
        nxpositioner["value"].attrs["units"] = "um"
        if amesh.slow.name not in measurement:
            measurement[amesh.slow.name] = h5py.SoftLink(nxpositioner["value"].name)
        if amesh.slow.name not in positioners:
            positioners[amesh.slow.name] = h5py.SoftLink(nxpositioner["value"].name)

        # MCA detector
        for i in range(nmcas):
            det_name = f"mca{i}"
            nxdetector = nxinstrument.require_group(det_name)
            nxdetector.attrs["NX_class"] = "NXdetector"
            for signal_name, signal_values in measured_data.items():
                if signal_name in nxdetector:
                    del nxdetector[signal_name]
                if signal_name == "spectrum":
                    mca_shape = (amesh.size, signal_values.shape[-1])
                    nxdetector[signal_name] = signal_values.reshape(
                        mca_shape, order="F"
                    )
                    if "data" not in nxdetector:
                        nxdetector["data"] = h5py.SoftLink("spectrum")
                    meas_name = det_name
                else:
                    nxdetector[signal_name] = signal_values.flatten(order="F")
                    meas_name = f"{det_name}_{signal_name}"
                if meas_name not in measurement:
                    measurement[meas_name] = h5py.SoftLink(nxdetector[signal_name].name)

        nxprocess = nxentry.require_group("theory")
        nxprocess.attrs["NX_class"] = "NXprocess"
        if "I0_reference" in nxprocess:
            del nxprocess["I0_reference"]
        nxprocess["I0_reference"] = I0_reference

        nxnote = nxprocess.require_group("configuration")
        nxnote.attrs["NX_class"] = "NXnote"
        if "data" in nxnote:
            del nxnote["data"]
        if "type" in nxnote:
            del nxnote["type"]
        nxnote["type"] = "application/pymca"
        nxnote["data"] = pymca_configdict_from_model(pymca_configuration).tostring()

        nxnote = nxprocess.require_group("description")
        nxnote.attrs["NX_class"] = "NXnote"
        if "data" in nxnote:
            del nxnote["data"]
        if "type" in nxnote:
            del nxnote["type"]
        nxnote["type"] = "text/plain"
        description = [
            "- parameters: peak areas without dead-time",
            "- parameters_norm: peak areas without dead-time and I0 normalized",
            "- rois: MCA ROI's without dead-time and I0 normalized (theoretical)",
            "- rois_norm: MCA ROI's without dead-time and I0 normalized (calculated)",
        ]
        nxnote["data"] = "\n".join(description)

        signals = {f"{g.element}-{g.name}": g.counts for g in linegroups}
        signals.update({g.name: g.counts for g in scattergroups})
        _save_nxdata(amesh, nxprocess, "parameters", signals, positioners)

        signals = {
            f"{g.element}-{g.name}": g.counts / I0 * I0_reference for g in linegroups
        }
        signals.update({g.name: g.counts / I0 * I0_reference for g in scattergroups})
        _save_nxdata(amesh, nxprocess, "parameters_norm", signals, positioners)

        if roi_data_theory:
            _save_nxdata(amesh, nxprocess, "rois", roi_data_theory, positioners)

        if roi_data_cor:
            _save_nxdata(amesh, nxprocess, "rois_norm", roi_data_cor, positioners)

        if "end_time" in nxentry:
            del nxentry["end_time"]
        nxentry["end_time"] = nexus_utils.now()


def _save_nxdata(
    amesh: _Amesh,
    parent: hdf5.GroupType,
    name: str,
    signals: Dict[str, numpy.ndarray],
    positioners: hdf5.GroupType,
) -> None:
    """Saves a set of signals as NeXus data to a given parent group.

    :param amesh: Scan description.
    :param parent: The parent `h5py.Group` where the data will be saved.
    :param name: The name of the dataset to be saved under.
    :param signals: Dictionary of signal names and their corresponding data arrays.
    :param positioners: Group containing the positioners (`sampz` and `sampy`) for the scan.
    """
    nxdata = parent.require_group(name)
    nxdata.attrs["NX_class"] = "NXdata"
    names = list(signals.keys())
    nxdata.attrs["signal"] = names[0]
    if len(names) > 1:
        nxdata.attrs["auxiliary_signals"] = names[1:]
    for signal_name, signal_values in signals.items():
        if signal_name in nxdata:
            del nxdata[signal_name]
        nxdata[signal_name] = signal_values.flatten(order="F")
    nxdata.attrs["axes"] = [amesh.fast.name, amesh.slow.name]  # Order: fast to slow
    if amesh.fast.name not in nxdata:
        nxdata[amesh.fast.name] = h5py.SoftLink(positioners[amesh.fast.name].name)
    if amesh.slow.name not in nxdata:
        nxdata[amesh.slow.name] = h5py.SoftLink(positioners[amesh.slow.name].name)


def _iter_amesh_data(
    amesh: _Amesh,
    coordinates: List[numpy.ndarray],
    rstate: numpy.random.RandomState,
    nmaps: int = 1,
) -> Iterator[numpy.ndarray]:
    """Yield random samples of an image scanned with `amesh`.

    :param amesh: Scan description.
    :param coordinates: Scan coordinates (fast axis first).

    :returns: signal (F-order matrix).
    """
    scene_data, scene_axes = _scene_data()
    scan_axes = [amesh.fast.name, amesh.slow.name]
    if scan_axes == scene_axes:
        pass
    elif scan_axes == scene_axes[::-1]:
        # Flip fast and slow axis
        scene_data = numpy.transpose(scene_data, [0, 2, 1])
    else:
        raise ValueError("Must be an amesh scan over 'sampy' and 'sampz'")

    flat_coordinates = [x.flatten(order="F") for x in coordinates]
    for _ in range(nmaps):
        # Random linear combination of the RGB channels which
        # results in an image with values between 0 and 1
        fractions = rstate.uniform(low=0, high=1, size=3)
        fractions /= 255 * fractions.sum()
        image = sum(fractions[:, numpy.newaxis, numpy.newaxis] * scene_data)

        # Interpolate the image (pixel coordinates) on the coordinate grid
        iimage = ndimage.map_coordinates(
            image, flat_coordinates, order=1, cval=0, mode="nearest"
        )
        yield iimage.reshape(amesh.shape, order="F")


def _amesh_positions(
    asmesh: _Amesh,
    rstate: numpy.random.RandomState,
    max_deviation: float = 0,
) -> List[numpy.ndarray]:
    """Generates motor positions for an amesh scan with optional deviations (fast axis first)."""
    positions = numpy.meshgrid(
        asmesh.fast.coordinates, asmesh.slow.coordinates, indexing="ij"
    )
    if not max_deviation:
        return positions
    deviations = [
        abs(
            max_deviation
            if axis.size <= 1
            else (axis.end - axis.start) / (axis.size - 1) * max_deviation
        )
        for axis in [asmesh.fast, asmesh.slow]
    ]
    positions = [
        values + rstate.uniform(low=-d, high=d, size=asmesh.shape)
        for values, d in zip(positions, deviations)
    ]
    return positions



import logging
from typing import Dict

from ..model import PyMcaXrfConfiguration
from .matrix import ensure_matrix

logger = logging.getLogger(__name__)


def enable_quantification(
    configuration: PyMcaXrfConfiguration, quantification: Dict[str, float]
):
    concentrations = configuration.concentrations
    attenuators = configuration.attenuator_views
    matrix = attenuators["Matrix"]

    if "flux" in quantification:
        concentrations.flux = quantification["flux"]
        logger.warning(
            "PyMca configuration: set flux to %e photons/s", concentrations.flux
        )

    if "time" in quantification:
        concentrations.time = quantification["time"]
        logger.warning(
            "PyMca configuration: set effective measurment time to %f s",
            concentrations.time,
        )

    if "area" in quantification:
        concentrations.area = quantification["area"]
        logger.warning(
            "PyMca configuration: set active detector area to %f cm^2",
            concentrations.area,
        )

    if "distance" in quantification:
        concentrations.distance = quantification["distance"]
        logger.warning(
            "PyMca configuration: set sample-detector distance to %f cm",
            concentrations.distance,
        )

    ensure_matrix(configuration)

    if "anglein" in quantification:
        matrix.incidence_angle = quantification["anglein"]
        logger.warning(
            "PyMca configuration: set incident angle to %f degrees",
            concentrations.incidence_angle,
        )

    if "angleout" in quantification:
        matrix.exit_angle = quantification["angleout"]
        logger.warning(
            "PyMca configuration: set exit angle to %f degrees",
            concentrations.exit_angle,
        )

    if "anglein" in quantification and "angleout" in quantification:
        matrix.scattering_angle = quantification["anglein"] + quantification["angleout"]
        logger.warning(
            "PyMca configuration: set scattering angle to %f degrees",
            concentrations.scattering_angle,
        )



from typing import Callable
from typing import Mapping
from typing import Optional

from AnyQt import QtWidgets
from ewokscore import missing_data
from ewoksorange.gui.widgets.parameter_form import ParameterForm

from . import serialize


class PickScansWidget(QtWidgets.QWidget):
    def __init__(
        self,
        default_inputs_change_callback: Callable,
        initial_values: Optional[Mapping] = None,
    ):
        super().__init__()

        if initial_values is None:
            initial_values = {}

        self.form = ParameterForm(parent=self)
        self._default_inputs_change_callback = default_inputs_change_callback

        parameters = {
            "filename": {
                "label": "HDF5 filename",
                "value_for_type": "",
                "select": "file",
            },
            "scan_min": {
                "label": "Minimum scan to include",
                "value_for_type": 1,
            },
            "scan_max": {
                "label": "Maximum scan to include",
                "value_for_type": 1,
            },
            "exclude_scans": {
                "label": "List of scans to exclude",
                "value_for_type": "",
                "serialize": serialize.integers_serializer,
                "deserialize": serialize.integers_deserializer,
            },
        }

        for name, kw in parameters.items():
            self.form.addParameter(
                name,
                value=initial_values.get(name, missing_data.MISSING_DATA),
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self) -> None:
        self._default_inputs_change_callback(self.form.get_parameter_values())



from typing import List
from typing import Optional

import h5py
import numpy
from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io import hdf5
from .. import hdf5_utils
from .. import nexus_utils
from ..models import FluoInputs


class Inputs(FluoInputs):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/merge",
        ],
    )

    xrf_results_uri: Optional[str] = Field(
        default=None,
        description="Previous XRF NXprocess URI to merge the raw Bliss counters with.",
        examples=["/results/dataset.h5::/1.1/sum/results"],
    )


class Outputs(BaseOutputModel):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/merge/results"],
    )


class ExtractRawCountersStack(Task, input_model=Inputs, output_model=Outputs):
    """Extract raw stack counters and save them like XRF results."""

    def run(self):
        start_time = nexus_utils.now()
        bliss_scan_uris = self.inputs.bliss_scan_uris
        output_root_uri = self.inputs.output_root_uri
        previous_xrf_results_uri = self.inputs.xrf_results_uri

        input_file0, scan_h5path0 = hdf5.split_h5uri(bliss_scan_uris[0])
        if previous_xrf_results_uri:
            default_nprocess_name = "merge"
        else:
            default_nprocess_name = "raw"

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config=dict(),
            default_levels=(
                scan_h5path0,  # TODO: what should the default be?
                default_nprocess_name,
            ),
        ) as (process_group, already_existed):
            if already_existed:
                merged_xrf_results = process_group["results"]
            else:
                merged_xrf_results = process_group.create_group("results")
                merged_xrf_results.attrs["NX_class"] = "NXcollection"

                with hdf5.FileReadAccess(input_file0) as h5file0:
                    _link_raw_counters(
                        h5file0[scan_h5path0],
                        bliss_scan_uris,
                        merged_xrf_results,
                        "rawcounters",
                    )

                if previous_xrf_results_uri:
                    input_file, parent_path = hdf5.split_h5uri(previous_xrf_results_uri)
                    with hdf5.FileReadAccess(input_file) as h5file:
                        _link_xrf_results(h5file[parent_path], merged_xrf_results)

            self.outputs.xrf_results_uri = (
                f"{merged_xrf_results.file.filename}::{merged_xrf_results.name}"
            )

        self.outputs.bliss_scan_uris = bliss_scan_uris
        self.outputs.output_root_uri = output_root_uri


def _link_raw_counters(
    raw_scan0: hdf5.GroupType,
    bliss_scan_uris: List[str],
    process_group: hdf5.GroupType,
    name: str,
) -> None:
    destination = nexus_utils.create_nxdata(process_group, name)
    measurement = raw_scan0["measurement"]

    nscans = len(bliss_scan_uris)
    layouts = dict()
    for name, dset in measurement.items():
        if dset.ndim == 1:
            layout = h5py.VirtualLayout(shape=(nscans, dset.size), dtype=dset.dtype)
            vsource_shape = (dset.size,)
            layouts[name] = layout, vsource_shape

    for i, bliss_scan_uri in enumerate(bliss_scan_uris):
        filename, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
        for name, (layout, vsource_shape) in layouts.items():
            layout[i] = h5py.VirtualSource(
                filename, f"{scan_h5path}/measurement/{name}", shape=vsource_shape
            )

    for name, (layout, _) in layouts.items():
        destination.create_virtual_dataset(name, layout, fillvalue=numpy.nan)

    nexus_utils.set_nxdata_signals(destination, signals=list(layouts))


def _link_xrf_results(
    xrf_results: hdf5.GroupType, process_group: hdf5.GroupType
) -> None:
    for name, group in xrf_results.items():
        hdf5_utils.create_hdf5_link(process_group, name, group)
